

# just in case, define macros for external utilities we use
PERL:=perl
ECHO:=echo
MKDIR:=mkdir
TOUCH:=touch
RM:=rm

ifndef ROOT_SRC
  ROOT_SRC:=$(ROOT_PROJECT)/src
  export ROOT_SRC
endif

GLOBIFY:=$(PERL) $(ROOT_SCORE)/bin/globify.pl

ifndef BUILD_MODE
  BUILD_MODE:=debugging
endif
ifndef PLATFORM
  PLATFORM:=linux
endif

export BUILD_MODE PLATFORM

ifndef ROOT_DERIVED
  ROOT_DERIVED:=$(ROOT_PROJECT)/derived
  export ROOT_DERIVED
endif

# HERE and derived variables must be recomputed on each invocation
# do NOT export them.
HERE:=$(shell pwd)
ifndef HERE_DASHED
  HERE_DASHED:=$(foreach dir,$(ROOT_SRC),$(subst $(dir)/,,$(HERE)/))
  HERE_DASHED:=$(filter-out $(HERE)/,$(HERE_DASHED))
  HERE_DASHED:=$(subst /,-,$(firstword $(HERE_DASHED)))
  HERE_DASHED:=$(MANGLE_PREFIX)$(HERE_DASHED)
endif

#
# Functions that compute the standard path variables given the derived
# directory name.  The standard path vars are as follows:
# LIBPATH   (eg /home/foo/src/root_score/derived/pc_release_dll/lib)
# LIBDIR    (eg /home/foo/src/root_score/derived/pc_release_dll/lib/mylib)
# EXEDIR    (eg /home/foo/src/root_score/derived/pc_release_dll/exe)
# _SLASH variants of all above (with /$(HERE_DASHED) tacked on)
#
# The intent is to call only set-path-variables once from
# the given platform.mk like this:
# $(eval $(call set-path-variables,$(DERIVED)))
#
define set-libpath
  ifndef LIBPATH
    LIBPATH:=$(ROOT_DERIVED)/$1/lib
  endif
endef
define set-libdir
  ifdef destLib
    LIBDIR:=$(LIBPATH)/$(destLib:%.lib=%)
  endif
endef
define set-exedir
  EXEDIR:=$(ROOT_DERIVED)/$1/exe
endef
define set-path-variables
  $(eval $(call set-libpath,$1))
  $(eval $(call set-libdir,$1))
  $(eval $(call set-exedir,$1))
  EXEDIR_SLASH:=$(EXEDIR)/$(HERE_DASHED)
  ifdef destLib
    LIBDIR_SLASH:=$(LIBDIR)/$(HERE_DASHED)
  endif
endef

override CFLAGS_CPP += -Dsm$(BUILD_MODE)
override CFLAGS_CPP += -Dsm_platform_$(PLATFORM)

CLEANABLE+=$(EXEDIR_SLASH)*
ifdef destLib
  CLEANABLE+=$(LIBDIR_SLASH)*
endif

ifdef QUERY
.PHONY: _default
_default:
	@echo $($(QUERY))
endif

# default target: recursively build here.
.PHONY: all
all:
	@RECURSE=: $(MAKE) -w $(MAKE_FLAGS) recurse

.PHONY: allhere
allhere: $(PLATFORM) ;

.PHONY: cleanhere
cleanhere:
	@$(RM) -f $(CLEANABLE) || true

.PHONY: libs
libs: $(PLATFORM:%=libs_%) ;

include $(ROOT_PROJECT)/build/recurse.mk

###############################################################################
# note: PLATFORM & TARGET makefiles must compute:
# GLOBJS (non-gnu-style object list for glob files)
# standard path variables as: $(eval $(call set-path-variables,$(DERIVED)))
# *all* build rules 
-include $(ROOT_PROJECT)/build/$(PLATFORM)/platform.mk


#
# Keep intermediate files so they dont keep getting rebuilt
#
.PRECIOUS: %.lib
.PRECIOUS: $(OTHER_OBJS_COMMON:%=$(EXEDIR_SLASH)%.obj)
.PRECIOUS: %.updated

%.updated:
	@$(TOUCH) $@

# automatic dependency generation
ifdef destLib
D_FILES+=\
 $(patsubst %.c,$(LIBDIR_SLASH)%.d,$(filter %.c,$(SRCS)))\
 $(patsubst %.cpp,$(LIBDIR_SLASH)%.d,$(filter %.cpp,$(SRCS)))
else
D_FILES+=\
 $(patsubst %.c,$(EXEDIR_SLASH)%.d,$(filter %.c,$(SRCS)))\
 $(patsubst %.cpp,$(EXEDIR_SLASH)%.d,$(filter %.cpp,$(SRCS)))
endif

ifndef INHIBIT_DEPS
  ifneq "$(filter clean,$(MAKECMDGOALS))" ""
    ifeq "$(CLEANTARGETCOMPLETE)" "1"
      -include $(D_FILES)
    endif
  else 
      -include $(D_FILES)
  endif
endif
CLEANABLE+=$(D_FILES)

# automatic library member tracking via destLib variable and glob files
GLOBFILE:=$(patsubst -%,%,$(subst /,-,$(HERE)).glob)
ifndef ROOTGLOB
  ROOTGLOB:=$(patsubst -%,%,$(subst /,-,$(ROOT_SRC)).glob)
  export ROOTGLOB
endif

ifneq "$(destLib)$(DIRS)" ""
$(LIBPATH)/$(GLOBFILE): $(HERE)/makefile
	@echo building $@ ; mkdir -p $(@D)
	@( $(ECHO) "#lib: $(destLib:%.lib=%) $(GLOBJS)"; $(ECHO) "#subdirs: $(DIRS:%=$(LIBPATH)/$(patsubst -%,%,$(subst :,,$(subst /,-,$(HERE))))-%.glob)" ) >$@
endif

ifndef INHIBIT_GLOBS
  ifneq "$(filter clean,$(MAKECMDGOALS))" ""
    ifeq "$(CLEANTARGETCOMPLETE)" "1"
      -include $(LIBPATH)/$(GLOBFILE)
    endif
  else 
    -include $(LIBPATH)/$(GLOBFILE)
  endif
endif
CLEANABLE+=$(LIBPATH)/$(GLOBFILE)

.PHONY: clean
clean:
	@RECURSE='$$(RM) -f $$(CLEANABLE) ; [ -f test/makefile ] && $$(MAKE) -w -C test $$(MAKE_FLAGS) cleanhere ; true' $(MAKE) -w -$(MAKE_FLAGS) INHIBIT_DEPS=1 INHIBIT_GLOBS=1 recurse
	@$(eval CLEANTARGETCOMPLETE:=1)

