
CC      := gcc
LD      := ld -v
AR      := ar
AS      := as
OC      := objcopy
OD      := objdump
CPP     := cpp

PERL:=perl
GLOBIFY:=$(PERL) $(ROOT_PROJECT)/build/globify.pl

# choose reasonable defaults for the following platform specific variables:
# BUILD_TYPE (static, dll)
# RTL_TYPE   (static, dll)
# RTL_MODE   (release, debugging)

ifndef BUILD_TYPE
  BUILD_TYPE:=static
endif
ifndef RTL_TYPE
  RTL_TYPE:=static
endif
ifndef RTL_MODE
  ifeq ($(BUILD_MODE),debugging)
    RTL_MODE:=debugging
  else
    RTL_MODE:=release
  endif
endif

ifndef ARCH
  ARCH:=x86_64
endif

#
# Set default value for asm configuration
#
ifndef BUILD_ASM
  BUILD_ASM:=true
endif

include $(ROOT_PROJECT)/build/linux/compilers.mk

# Set up the name of the derived tree for the build configuration
DERIVED:=$(PLATFORM)_$(BUILD_MODE)_$(BUILD_TYPE)
ifneq ($(RTL_TYPE),dll)
  DERIVED:=$(DERIVED)_$(RTL_TYPE)
endif
ifeq ($(CHAR_MODE),unicode)
  DERIVED:=$(DERIVED)_u
endif
ifeq ($(RTL_MODE),debugging)
#  DERIVED:=$(DERIVED)_d
endif
ifdef COMPILER_SUFFIX
  DERIVED:=$(DERIVED)_$(COMPILER_SUFFIX)
endif

ifndef MINGW
  export MINGW=false
endif

# setup all path variables (LIBPATH, LIBDIR, EXEDIR, etc)
$(eval $(call set-path-variables,$(DERIVED)))

#
# Collect up sources
#
SRCS := $(SOURCES) $(SOURCES_LINUX) $(SOURCES_COMMON)
TSTS := $(TESTS) $(TESTS_LINUX) $(TESTS_COMMON)
EXES := $(EXES) $(EXES_LINUX) $(EXES_COMMON)
OBJS := $(patsubst %.s,%.o,$(patsubst %.cpp,%.o,$(patsubst %.c,%.o,$(SRCS))))

# The non-gnu-style versions of the OBJS_... names go into the .glob files.
GLOBJS:=$(OBJS:%=$(LIBDIR_SLASH)%)

ifneq ($(destLib),)
  OBJS:=$(OBJS:%=$(LIBDIR_SLASH)%)
else 
  OBJS:=$(OBJS:%=$(EXEDIR_SLASH)%)
endif

LIBDIRS=$(LIBPATH)

INC_DIRS=$(patsubst %,-I%,$(INCDIRS) $(INCLUDE_DIRS))
LIB_DIRS=$(patsubst %,-WL%,$(LIBDIRS) $(LIBRARY_DIRS))

CLEANABLE+= $(OBJS) $(EXES) $(HEXS)

ifdef destLib
  CLEANABLE+=$(LIBPATH)/$(destLib:%.lib=%.a)
endif

override CFLAGS_CPP += -DsmArch_$(ARCH) 
override CFLAGS_CPP+=$(INCLUDE_DIRS:%=-I%)

GENDEPY_LNX=$(CPP) -w -undef -U__STDC__ $(CFLAGS_CPP) $(CFLAG_DEFS_PLAT) $(CFLAGS_PLAT)
COMPILE_LNX=$(CC) $(CFLAGS_PLAT) $(CFLAGS_CPP) $(CFLAG_DEFS_PLAT) 
ASSMBLE_LNX=$(CC) $(CFLAGS_PLAT) $(CFLAGS_CPP) $(CFLAG_DEFS_PLAT) 

override CFLAG_DEFS_PLAT += -DsmLink_$(BUILD_TYPE) 

ifeq ($(BUILD_MODE),release)
  override CFLAGS_CPP += -DNDEBUG -O3
endif
ifeq ($(BUILD_MODE),debugging)
  override CFLAGS_CPP += -D_DEBUG
  CFLAGS_PLAT=-g -O0
endif
ifeq ($(ARCH),x86_64)
  CFLAGS_PLAT+=-m64
  override CFLAG_DEFS_PLAT+=-D__x86_64__
else 
  CFLAGS_PLAT+=-m32
endif
ifeq ($(MINGW),true)
  CFLAGS_PLAT=-isystem /usr/include/mingw
endif

#
# Handle consequences of destLib
#
ifdef destLib
  override CFLAGS_CPP += -DsmExport_$(destLib:%.lib=%) -DsmLibraryName=$(destLib:%.lib=%)
endif

export RTL_MODE
export RTL_TYPE
export BUILD_TYPE

ifndef LIBS_LINUX
  LIBS_LINUX:=$(LIBS_COMMON:%.lib=%.a)
  LIBS_LINUX:=$(shell perl -e "@a = reverse @ARGV; $$,=' '; print @a;" $(LIBS_LINUX))
# $(warning LIBS_LINUX -> $(LIBS_LINUX))
  export LIBS_LINUX
endif

define list-objs
$(patsubst %,$(EXEDIR_SLASH)%.o,$($1_$2))
endef

#define list-deps
#$(warning $1 dependencies -> $($(1)_OBJS_COMMON:%=$(EXEDIR_SLASH)%.o) $($(1)_OBJS_LINUX:%=$(EXEDIR_SLASH)%.o) $(EXTRA_LIBS) $(EXTRA_STATIC_LIBS:%.a=$(LIBPATH)/%.a) $(LIBS_LINUX:%=$(LIBPATH)/%))
#endef

define emit-exe-rule
$(EXEDIR_SLASH)$1 : $($(1)_OBJS_COMMON:%=$(EXEDIR_SLASH)%.o) $($(1)_OBJS_LINUX:%=$(EXEDIR_SLASH)%.o) $(EXTRA_LIBS) $(EXTRA_STATIC_LIBS:%.a=$(LIBPATH)/%.a) $(LIBS_LINUX:%=$(LIBPATH)/%)
	gcc -o $(EXEDIR_SLASH)$1 $(LIB_DIRS) $(call list-objs,$1,OBJS_LINUX) $(call list-objs,$1,OBJS_COMMON) $(LIBS) $(LIBS_LINUX:%=$(LIBPATH)/%)  $(EXTRA_LIBS) -lpthread -lm -lc

endef


#
# platform targets.
#
.PHONY: linux
ifndef INHIBIT_DEPS
linux: $(OBJS) $(OTHER_OBJS) $(EXES:%=$(EXEDIR_SLASH)%) $(TSTS:%=$(EXEDIR_SLASH)%) ;
else
linux: ;
endif

.PHONY: libs_linux
libs_linux: $(LIBS_LINUX:%.a=$(LIBPATH)/%.a) \
            $(LIBS_LINUX_OTHER_TARGETS) ;

.PHONY: run_linux
run_linux:
	@for t in $(TSTS:%=$(EXEDIR_SLASH)%); do \
     if [ -f $$t ]; then                       \
       $$t ;                                   \
     else                                      \
       echo "skipping test $$d" ;              \
     fi;                                       \
	 done                                        

.PHONY: list
list:
	@echo $(SRCS)

$(LIBDIR_SLASH)%.d $(EXEDIR_SLASH)%.d : %.c  
	@echo building $@ ;                                                      \
	mkdir -p $(@D) &&                                                        \
	$(GENDEPY_LNX) -MM $< > $@.$$$$;                                         \
	sed 's|\($*\)\.o[ ]*:|$(@D)\/$(HERE_DASHED)\1.o $@ : |g' < $@.$$$$ > $@; \
	rm -f $@.$$$$; 

$(LIBDIR_SLASH)%.d $(EXEDIR_SLASH)%.d: %.cpp
	@echo building $@ ;                                                      \
	mkdir -p $(@D) &&                                                        \
	$(GENDEPY_LNX) -D__cplusplus -MM $< > $@.$$$$;                           \
	sed 's|\($*\)\.o[ ]*:|$(@D)\/$(HERE_DASHED)\1.o $@ : |g' < $@.$$$$ > $@; \
	rm -f $@.$$$$; 

# Rules for generating object files
$(LIBDIR_SLASH)%.o : %.c
	$(COMPILE_LNX) -c -o $(LIBDIR)/$(@F) $(INC_DIRS) $<
	@$(TOUCH) $(LIBPATH)/$(destLib:%.lib=%).updated

$(EXEDIR_SLASH)%.o : %.c
	$(COMPILE_LNX) -c -o $(EXEDIR)/$(@F) $(INC_DIRS) $<

$(LIBDIR_SLASH)%.o : %.s
	$(ASSMBLE_LNX) -c -o $(LIBDIR)/$(@F) $< > $(LIBDIR)/$(@F).lst
	@$(TOUCH) $(LIBPATH)/$(destLib:%.lib=%).updated

$(EXEDIR_SLASH)%.o : %.s
	$(ASSMBLE_LNX) -c -o $(EXEDIR)/$(@F) $< > $(EXEDIR)/$(@F).lst

$(EXEDIR_SLASH)%.o : %.cpp
	$(COMPILE_LNX) -c -o $(EXEDIR)/$(@F) $(INC_DIRS) $<

#
# Rules for building static libraries
#   - using a temp file for the contents of the globify to avoid command line
#     length problems
_T_LINK=$(LIBPATH)/templink~

$(LIBPATH)/%.a: $(LIBPATH)/%.updated
	@$(RM) -f $@ ;                                                              \
	$(GLOBIFY) $(subst .a,,$(notdir $@)) $(LIBPATH)/$(ROOTGLOB) > $(_T_LINK) ;  \
	$(AR) -r $@ `cat $(_T_LINK)`                                       

#
# Rules for building listed executables, auto generated
ifndef INHIBIT_DEPS
  $(eval $(foreach exe,$(basename $(EXES)),$(call emit-exe-rule,$(exe))))
  $(eval $(foreach tst,$(basename $(TSTS)),$(call emit-exe-rule,$(tst))))
endif

#$(eval $(foreach tst,$(basename $(TSTS)),$(call list-deps,$(tst))))
