#ifndef __si_sha_autodev_h__fa6d367f_626d_4990_90b5_32aa069f1145
#define __si_sha_autodev_h__fa6d367f_626d_4990_90b5_32aa069f1145

/** \file si_sha/autodev.h
 *  \brief A base representation of automation devices in LASE.
 *
 *  All devices in LASE must have a name, address, type, and netType.
 *  If some devices lack such standards, they can be assigned by LASE.
 *  
 */

#include <si_std/string.h>
#include <si_std/listt.h>
#include <si_std/ptrh.h>

typedef struct sc_automationBridgeDevice sc_automationBridgeDevice;

/** \brief A base representation of all devices in LASE.
 *
 *  All devices in LASE must have a name, address, type, and netType.
 *  If some devices lack such standards, they can be assigned by LASE.
 *  
 */
typedef struct sc_automationDevice {
  sc_rep  _rep;          /**< Devices may be pointer-handled. */

  const TCHAR* name_;    /**< Identifier of the device in LASE. */
  const TCHAR* type_;    /**< Type of device. */
  const TCHAR* address_; /**< Address of the device on its network. */
  const TCHAR* netType_; /**< Identifies the network type. */

  sc_automationBridgeDevice* brg_; /**< Bridge device is attached to. */
  
} sc_automationDevice;

sc_status sc_automationDevice_construct(sc_automationDevice* pd,
                                        const TCHAR* name,
                                        const TCHAR* type,
                                        const TCHAR* addr,
                                        const TCHAR* netType);
/**< Constructs a device with the given name, type, address and network
 *   type.
 *   
 * \param[in,out] pd    Device object to construct.
 * \param[in]     name  The identifier the device uses.
 * \param[in]     type  The device type.
 * \param[in]     addr  The address of the device on its network.
 * \param[in]     netType The network type the device is connected to.
 *   
 */

sc_status sc_automationDevice_destruct(sc_automationDevice* pd);
/**< Destructs the device base class.
 */

/** \brief Macro declared sc_automationDevicePtrh: device
 * pointer-handles.
 */
sm_ptrhDecl(sc_automationDevice);

sc_status sc_automationDevice_new(const TCHAR* name,
                                  const TCHAR* type,
                                  const TCHAR* addr,
                                  const TCHAR* ntyp,
                                  sc_automationDevicePtrh* ptrh);
/**< Creates a new base device and returns it in \p ptrh.
 * \see sc_automationDevice_construct
 */

#endif /* __si_sha_autodev_h__fa6d367f_626d_4990_90b5_32aa069f1145 */

