#ifndef __si_sha_zigbee_h__292fe11d_e194_4a6d_8758_a8f67831b30e
#define __si_sha_zigbee_h__292fe11d_e194_4a6d_8758_a8f67831b30e

/** \file si_sha/zigbee.h
 *  \brief ZigBee technology data structures, constants and utilities.
 */

/** \brief NetId for ZigBee.
 *
 * Given by the LASE system to distinguish
 * it from other automation networks.
 */
#define ZIGBEE_NETID 0x02


/** \brief Macro to print zigbee addresses to a character buffer.
 *
 *  27 characters are required in buffer to hold the address. 
 */
#define zb_aprint(b,a)\
    sprintf(b,"%02X-%02X-%02X-%02X-%02X-%02X-%02X-%02X-%02X",\
            (int)ZIGBEE_NETID,\
            (int)(((a) >> 56) & 0xFF),\
            (int)(((a) >> 48) & 0xFF),\
            (int)(((a) >> 40) & 0xFF),\
            (int)(((a) >> 32) & 0xFF),\
            (int)(((a) >> 24) & 0xFF),\
            (int)(((a) >> 16) & 0xFF),\
            (int)(((a) >> 8)  & 0xFF),\
            (int)(((a) >> 0)  & 0xFF))

/** \brief Macro to print zigbee types to a character buffer.
 *
 *  9 characters are required in buffer to hold the type. 
 */
#define zb_tprint(b,t) \
    sprintf(b,"%02X-%02X-%02X",ZIGBEE_NETID,(int)(((t)>>8) & 0xFF),(int)((t) & 0xFF))


/** \name Remote command indexes.
 *
 *  Remote commands are a shorthand between the JavaScript side device
 *  and the bridge side. 
 */
/*@{*/
#define ZC_DISCOVER    1
#define ZC_STDMESSAGE  2
#define ZC_EXTMESSAGE  3
#define ZC_GRPMESSAGE  4
#define ZC_GETVERSION  5
#define ZC_FORMNETWORK 6
/*@}*/

/** \brief Type assigned to Atmel's RAVEN bridge device. */
#define ZB_RAVENBRGT   (0x00020001)

/** \brief ZigBee Bridge version information.
 *
 *  Contains information about the bridge device such as address and
 *  type.
 */
typedef struct sc_zigbeeBridgeVersionInfo {

  s_uint64 address_; /**< Address of device on network (64-bit IEEE address). */
  s_uint32 type_;    /**< Device type */
  
} sc_zigbeeBridgeVersionInfo;


#endif /* __si_sha_zigbee_h__292fe11d_e194_4a6d_8758_a8f67831b30e */
