#ifndef __si_std_h_uuid_h__8847b706_9ec0_4c88_a029_326f34f9cf9f
#define __si_std_h_uuid_h__8847b706_9ec0_4c88_a029_326f34f9cf9f

/*! \file si_std/uuid.h
 *  \brief Basic facility for uuid/guids
 *
 * Used here so that device categories and interfaces
 * can be independently developed without need to cooperate
 * on id constants or ranges.
 *
 * \note Only minimal support is included here.
 * 
 */


#include <si_sys/defs.h>


/*! \brief A UUID structure.
 */
typedef struct sc_uuid {  
  s_uint8 data[16];  /**< Storage for the UUID data. Opaque. */
} sc_uuid;

sc_status sc_uuid_init(sc_uuid* uuid);
/**< initialize the uuid from another uuid (may be NULL). */

sc_status sc_uuid_initFromString(sc_uuid* uuid, const TCHAR* uuidstr);
/**< Initialize the uuid from a uuid string given in the format:
 *   XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
 *                 or
 *   { XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX }
 *   \param[in,out] uuid UUID object to initialize.
 *   \param[in] uuidstr UUID string to convert.
 *   \note UUID strings can be obtained on the command line with most
 *   systems using the 'uuidgen' utility.
 */

sc_status sc_uuid_assign(sc_uuid* pthis, const sc_uuid* pthat);
/**< Assign pthis to the value of *pthat.
 */

sc_status sc_uuid_compareEqual(const sc_uuid* g0, const sc_uuid* g1,
                               s_bool* equal);
/**< Compares g0 to g1, and returns seTrue if they are equal, and
 * seFalse otherwise.
 * \param[in] g0 A UUID
 * \param[in] g1 Another UUID
 * \param[out] equal Gets set to seTrue if they are equal, seFalse
 *                   otherwise.
 */



#endif /* __si_std_h_uuid_h__8847b706_9ec0_4c88_a029_326f34f9cf9f */
