#ifndef __si_sys_defs_h__68cde394_4e51_4317_a76c_fa534c8082ef
#define __si_sys_defs_h__68cde394_4e51_4317_a76c_fa534c8082ef

/*! \file si_sys/defs.h
 *  \brief A file of standard typedefs used to make porting code
 * easier, and APIs more consistent.
 *
 */

#include <stdlib.h>

/**
 *  \name Simple integer type portability 
 */
/*@{*/
typedef unsigned char        s_uint8;
typedef char                 s_int8;
typedef unsigned short       s_uint16;
typedef short                s_int16;
typedef unsigned int         s_uint32;
typedef int                  s_int32;
typedef long long            s_int64;
typedef unsigned long long   s_uint64;

typedef enum { seFalse=0, seTrue } s_bool;

typedef char                 TCHAR;
#define _T(s) s
/*@}*/


#define sm_max(a,b) (((a)>(b))?(a):(b))


/*! \name Simple error handling facilities
 */
/*@{*/
typedef int sc_status;

#define sc_status_success   0
#define sc_status_failure  -1

#define sc_err_outOfMemory     (0xF0000001)
#define sc_err_badArgument     (0xF0000002)
#define sc_err_nullArgument    (0xF0000003)
#define sc_err_invalidIterator (0xF0000004)

#define sm_status_failed(s) ((s) & 0x80000000)
#define sm_status_succeeded(s) (!sm_status_failed(s))


#ifdef smrelease
#define dbgbreak(s) ;
#else
void dbgbreak(sc_status s);
#endif

#define sm_status_returnIfFailed(s) do { \
                                      if(sm_status_failed(s)) \
                                        return (s);\
                                    } while(0)

#define sm_status_returnError(s,e) do { \
                                     (s) = (e); \
                                     dbgbreak(s);\
                                     return (s);\
                                   } while(0)
/**< Returns error e in status s.
 * \note In debug mode, the initial line where an error was generated
 * calls the function dbgbreak().  Try setting a break point in this
 * function and then looking back up the call stack to find the line where
 * the error originated.
 */

#define sm_status_return(s) return s

/*@}*/

#endif /* __si_sys_defs_h__68cde394_4e51_4317_a76c_fa534c8082ef */


