#include <stdlib.h>
#include <stdio.h>
#include <limits.h>


#include <si_sha/instbdev.h>
#include <si_sha/insteon.h>
#include <si_std/string.h>

#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>


sc_status sc_insteonBridgeDevice_discover(sc_insteonBridgeDevice* pb);

sc_status sc_insteonBridgeDevice_construct(sc_insteonBridgeDevice* pbd,
                                           sc_automationDeviceMgr* pdm,
                                           const TCHAR* name,
                                           const TCHAR* bridgePath)
{
  static const sc_automationBridgeDevice_vtable vtbl = {  /* derived v-table */
    (sc_automationBridgeDevice_processEventF)sc_insteonBridgeDevice_processEvent,
    (sc_automationBridgeDevice_processRequestF)sc_insteonBridgeDevice_processRequest,
    (sc_automationBridgeDevice_newPacketF)sc_insteonBridgeDevice_newPacket,
    (sc_automationBridgeDevice_discoverF)sc_insteonBridgeDevice_discover
  };
  sc_status status = sc_status_success;
  sc_automationPacketPtrh rqptrh, rsptrh;
  sc_automationPacket *req, *res;
  sc_string namePath;
  sc_insteonBridgeVersion vs;
  s_uint16 result;
  char dtype[16] = {0};
  char daddr[16] = {0};
  struct sockaddr_un addr;
  int rv;

  /* construct pool and allocate a couple of packets */
  status = sc_automationPacketPool_construct(&pbd->pool_, 32, 16, INT_MAX);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacketPtrh_construct(&rqptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&rsptrh, NULL);
  sm_status_returnIfFailed(status);
  
  status = sc_automationPacketPool_allocate(&pbd->pool_, &rqptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPool_allocate(&pbd->pool_, &rsptrh);
  sm_status_returnIfFailed(status);
  req = sc_automationPacketPtrh_get(&rqptrh);
  res = sc_automationPacketPtrh_get(&rsptrh);

  /* build up bridge name path */
  sc_string_construct(&namePath);
  sc_string_append(&namePath, bridgePath);
  sc_string_appendChars(&namePath, "/", 1);
  sc_string_append(&namePath, name);  

  /* connect to bridge at bridgePath/name */
  pbd->base_.fd_ = socket(AF_UNIX, SOCK_STREAM, 0);
  
  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, namePath.str_, sizeof(addr.sun_path)-1);
  
  rv = connect(pbd->base_.fd_, (struct sockaddr*)&addr, sizeof(addr));
  if(rv != 0) {
    sc_automationPacketPtrh_destruct(&rsptrh);
    sc_automationPacketPtrh_destruct(&rqptrh);
    sc_string_destruct(&namePath);
    sc_automationPacketPool_destruct(&pbd->pool_);
    sm_status_returnError(status, sc_status_failure);
  }
  /* rv == 0 => success */

  status = sc_automationPacket_pushUInt16(req, RC_GETVERSION);
  sm_status_returnIfFailed(status);
  
  /* send a message to get the address */
  status = sc_insteonBridgeDevice_processRequest(pbd, req, res);
  sm_status_returnIfFailed(status);

  status = sc_automationPacket_popUInt16(res, &result);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_popUInt32(res, &vs.address_);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_popUInt8(res, &vs.category_);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_popUInt8(res, &vs.subcategory_);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_popUInt8(res, &vs.firmwareVersion_);
  sm_status_returnIfFailed(status);

  /* print address and type strings */
  tprint(dtype, vs.category_, vs.subcategory_);
  aprint(daddr, vs.address_);
  
  status = sc_automationBridgeDevice_construct(&pbd->base_, pdm, name,
                                               dtype, daddr, "Insteon",
                                               &vtbl, pbd);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPtrh_destruct(&rqptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rsptrh);
  sm_status_returnIfFailed(status);
  status = sc_string_destruct(&namePath);
  sm_status_returnIfFailed(status);

  printf("Insteon bridge device type %s, address %s, connected.\n", dtype, daddr);

  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_newPacket(sc_insteonBridgeDevice* pb,
                                           sc_automationPacketPtrh* pptrh)
{
  return sc_automationPacketPool_allocate(&pb->pool_, pptrh);
}

sc_status sc_insteonBridgeDevice_processEvent(sc_insteonBridgeDevice* brg,
                                              sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  int fd;
  s_uint8 tmp, code, cmd1, cmd2;
  s_uint32 from, to, ppos;
  s_uint16 result;
  char fromAddress[16];

  fd = brg->base_.fd_;

  status = sc_automationPacket_deserializeFromFd(pkt, fd);
  sm_status_returnIfFailed(status);
  
  sc_automationPacket_tell(pkt, &ppos);
  sc_automationPacket_popUInt16(pkt, &result);  /* 0x0001 on success */
  sc_automationPacket_popUInt8 (pkt, &tmp);     /* 0x02 */
  sc_automationPacket_popUInt8 (pkt, &code);    /* event code */

  switch(code) {
    case se_insteonCode_stdInEvent: /* standard insteon message */
      sc_automationPacket_popUInt32(pkt, &from);
      sc_automationPacket_popUInt32(pkt, &to);
      sc_automationPacket_popUInt8 (pkt, &tmp); /* flags */
      sc_automationPacket_popUInt8 (pkt, &cmd1); 
      sc_automationPacket_popUInt8 (pkt, &cmd2);
      sc_automationPacket_setPos(pkt, ppos);
      aprint(fromAddress, from);
      sc_automationDeviceMgr_relayMessage(brg->base_.pdm_,
                                          fromAddress, pkt);
      
      printf("Event: Std Insteon from %X, to %X, command %X-%X\n",
             from, to, (int)cmd1, (int)cmd2);
      break;
    case se_insteonCode_buttonEvent: /* buttonEvent */
      sc_automationPacket_popUInt8 (pkt, &tmp); /* flags */    
      printf("Event: Button Event > %d\n", tmp);
      break;
    default:
      printf("Event: %X\n", code);
      break;      
  }
  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_destruct(sc_insteonBridgeDevice* pbd)
{
  sc_status status = sc_status_success;
  
  printf("Insteon bridge device %s, address %s, disconnecting.\n",
         pbd->base_.base_.name_, pbd->base_.base_.address_);
  
  status = sc_automationBridgeDevice_destruct(&pbd->base_);
  sm_status_returnIfFailed(status);

  status = sc_automationPacketPool_destruct(&pbd->pool_);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_processRequest(sc_insteonBridgeDevice* brg,
                                                sc_automationRequest* req,
                                                sc_automationResponse* res)
{
  sc_status status = sc_status_success;
  fd_set rfds;
  int nfds, rv, fd;
  struct timeval tv;

  tv.tv_sec = 10;
  tv.tv_usec = 0;

  fd = brg->base_.fd_;
  nfds = fd + 1;

  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);

  /* serialize request out */
  status = sc_automationPacket_serializeToFd(req, fd);
  sm_status_returnIfFailed(status);

  /* wait for response */
  rv = select(nfds, &rfds, NULL, NULL, &tv);
  if(rv != 1) {
    sm_status_returnError(status, sc_status_failure);
  }

  /* deserialize response from fd */
  status = sc_automationPacket_deserializeFromFd(res, fd);
  sm_status_returnIfFailed(status);

  /* TODO: check for request/response match, queue as event if not */
  
  sm_status_return(status);
}


static sc_status deviceFromLinkEntry(sc_automationDeviceMgr* pom,
                              sc_insteonLinkEntry* ple,
                              sc_automationDevicePtrh* pd)
{
  sc_status status = sc_status_success;
  char type[16];
  char addr[16];
  char *instName = NULL;
  s_uint32 uaddr;

  uaddr = ple->address_;
  
  tprint(type,ple->category_,ple->subcategory_);
  aprint(addr,uaddr);
  
  status = sc_automationDevice_new(NULL,type,addr,"Insteon", pd);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_discover(sc_insteonBridgeDevice* pb)
{ 
  sc_status status = sc_status_success;
  sc_automationDevicePtrh dev;
  sc_automationDevice* pd;
  sc_automationPacketPtrh ptrh;
  sc_automationPacket* pkt;
  sc_insteonLinkEntry le;
  int fd, nfds, rv;
  fd_set rfds;
  struct timeval tv;
  s_uint16 rc;
  void* jsiname = NULL;

  tv.tv_sec = 10;
  tv.tv_usec = 0;
  fd = pb->base_.fd_;
  nfds = fd + 1;

  /* discover uses custom comm loop */
  
  status = sc_automationPacketPtrh_construct(&ptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationDevicePtrh_construct(&dev, NULL);
  sm_status_returnIfFailed(status);

  status = sc_insteonBridgeDevice_newPacket(pb, &ptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&ptrh);

  /* send discover request to driver */
  status = sc_automationPacket_pushUInt16(pkt, RC_DISCOVER);
  sm_status_returnIfFailed(status);

  status = sc_automationPacket_serializeToFd(pkt, fd);
  sm_status_returnIfFailed(status);

  /* already sent, clear out for re-use */
  sc_automationPacket_clear(pkt);

  /* wait for response */
  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  rv = select(nfds, &rfds, NULL, NULL, &tv);
  if(rv != 1) {
    sm_status_returnError(status, sc_status_failure);
  }
  
  /* deserialize response from fd */
  status = sc_automationPacket_deserializeFromFd(pkt, fd);
  sm_status_returnIfFailed(status);

  status = sc_automationPacket_popUInt16(pkt, &rc);
  sm_status_returnIfFailed(status);

  le.isEOL = (s_uint8)((rc == 0) ? 1 : 0);
  
  while(!le.isEOL) {
    /* unpack the link entry */
    status = sc_automationPacket_popUInt8 (pkt, &le.firmwareVersion_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_popUInt8 (pkt, &le.subcategory_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_popUInt8 (pkt, &le.category_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_popUInt8 (pkt, &le.group_);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_popUInt8 (pkt, &le.flags_);
    sm_status_returnIfFailed(status);    
    status = sc_automationPacket_popUInt32(pkt, &le.address_);
    sm_status_returnIfFailed(status);
    
    status = deviceFromLinkEntry(pb->base_.pdm_, &le, &dev);
    sm_status_returnIfFailed(status);
    
    pd = sc_automationDevicePtrh_get(&dev);
    pd->brg_ = &pb->base_;

    /* see if we've already got one of these... */
    status = sc_automationDeviceMgr_getInstanceFromDeviceAddress(pb->base_.pdm_,
                                                                 pd->address_,
                                                                 &jsiname);
    if(jsiname == NULL) {
      /* no instance created yet, create one on global object now */
      sc_automationDeviceMgr_createDevice(pb->base_.pdm_, pb->base_.base_.name_, pd);
    }

    FD_ZERO(&rfds);
    FD_SET(fd, &rfds);    
    sc_automationPacket_clear(pkt);
    /* wait for next response */
    rv = select(nfds, &rfds, NULL, NULL, &tv);
    if(rv != 1) {
      sm_status_returnError(status, sc_status_failure);
    }

    /* deserialize response from fd, check end-of-list condition */
    status = sc_automationPacket_deserializeFromFd(pkt, fd);
    sm_status_returnIfFailed(status);
    status = sc_automationPacket_popUInt16(pkt, &rc);
    sm_status_returnIfFailed(status);
    le.isEOL = (s_uint8)((rc == 0) ? 1 : 0);
  }
  
  status = sc_automationDevicePtrh_destruct(&dev);
  sm_status_returnIfFailed(status);
  
  status = sc_automationPacketPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_sendStdMessage(sc_insteonBridgeDevice* pb,
                                                const TCHAR* address,
                                                s_uint8 cmd0, s_uint8 cmd1)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrh ptrh, rptrh;
  sc_automationPacket* pkt = NULL;
  s_uint8 ahb, amb, alb;
  s_uint32 addr;
  s_uint16 result;

  /* This is so common that there's a special function to make sending
   * 'standard' message as simple as possible. */
  
  status = sc_automationPacketPtrh_construct(&ptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&rptrh, NULL);
  sm_status_returnIfFailed(status);

  status = sc_insteonBridgeDevice_newPacket(pb, &ptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&ptrh);

  /* scan address out */
  sscanf(address, "01-%2hhx-%2hhx-%2hhx", &ahb, &amb, &alb);
  addr = atou32(ahb,amb,alb);

  /* pack up the data and serialize to driver */
  status = sc_automationPacket_pushUInt8 (pkt, cmd1);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_pushUInt8 (pkt, cmd0);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt32(pkt, addr);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt16(pkt, RC_STDMESSAGE);
  sm_status_returnIfFailed(status);  

  status = sc_automationBridgeDevice_sendMessage(&pb->base_, pkt, &rptrh);
  sm_status_returnIfFailed(status);

  pkt = sc_automationPacketPtrh_get(&rptrh);
  if(pkt == NULL) {
    sc_automationPacketPtrh_destruct(&ptrh);
    sm_status_returnError(status, sc_status_failure);
  }

  /* get result back for status */
  status = sc_automationPacket_popUInt16(pkt, &result);
  sm_status_returnIfFailed(status);
  if(result != 0) {
    sc_automationPacketPtrh_destruct(&ptrh);
    sc_automationPacketPtrh_destruct(&rptrh);
    sm_status_returnError(status, sc_status_failure);    
  }
  
  status = sc_automationPacketPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_sendStdWaitAck(sc_insteonBridgeDevice* pb,
                                                const TCHAR* address,
                                                s_uint8 cmd0, s_uint8 cmd1,
                                                sc_automationPacketPtrh* rph)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrh ptrh, rptrh;
  sc_automationPacket* pkt = NULL;
  s_uint8 ahb, amb, alb;
  s_uint32 addr;
  s_uint16 result;

  status = sc_automationPacketPtrh_construct(&ptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&rptrh, NULL);
  sm_status_returnIfFailed(status);

  status = sc_insteonBridgeDevice_newPacket(pb, &ptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&ptrh);

  /* scan address out */
  sscanf(address, "01-%2hhx-%2hhx-%2hhx", &ahb, &amb, &alb);
  addr = atou32(ahb,amb,alb);

  /* pack it up ... wait is on other side */
  status = sc_automationPacket_pushUInt8 (pkt, cmd1);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacket_pushUInt8 (pkt, cmd0);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt32(pkt, addr);
  sm_status_returnIfFailed(status);
  status = sc_automationPacket_pushUInt16(pkt, RC_STDMSGWACK);
  sm_status_returnIfFailed(status);  

  status = sc_automationBridgeDevice_sendMessage(&pb->base_, pkt, &rptrh);
  sm_status_returnIfFailed(status);

  pkt = sc_automationPacketPtrh_get(&rptrh);
  if(pkt == NULL) {
    sc_automationPacketPtrh_destruct(&ptrh);
    sm_status_returnError(status, sc_status_failure);
  }

  /* check result */
  status = sc_automationPacket_popUInt16(pkt, &result);
  sm_status_returnIfFailed(status);
  if(result & 0x8000) { 
    sc_automationPacketPtrh_destruct(&ptrh);
    sc_automationPacketPtrh_destruct(&rptrh);
    sm_status_returnError(status, sc_status_failure);    
  }

  /* Assign the data back out */
  if(rph) {
    status = sc_automationPacket_pushUInt16(pkt, 0);
    sm_status_returnIfFailed(status);
    status = sc_automationPacketPtrh_assign(rph, &rptrh);
    sm_status_returnIfFailed(status);
  }
  
  status = sc_automationPacketPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sc_status sc_insteonBridgeDevice_resetDeviceDB(sc_insteonBridgeDevice* pb)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrh ptrh, rptrh;
  sc_automationPacket* pkt = NULL;
  s_uint16 result;
  
  status = sc_automationPacketPtrh_construct(&ptrh, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&rptrh, NULL);
  sm_status_returnIfFailed(status);

  /* allocate packet */
  status = sc_insteonBridgeDevice_newPacket(pb, &ptrh);
  sm_status_returnIfFailed(status);
  pkt = sc_automationPacketPtrh_get(&ptrh);

  status = sc_automationPacket_pushUInt16(pkt, RC_RESETLINKS);
  sm_status_returnIfFailed(status);  

  status = sc_automationBridgeDevice_sendMessage(&pb->base_, pkt, &rptrh);
  sm_status_returnIfFailed(status);

  pkt = sc_automationPacketPtrh_get(&rptrh);
  if(pkt == NULL) {
    sc_automationPacketPtrh_destruct(&ptrh);
    sm_status_returnError(status, sc_status_failure);
  }

  /* check result for status */
  status = sc_automationPacket_popUInt16(pkt, &result);
  sm_status_returnIfFailed(status);
  if(result != 0) {
    sc_automationPacketPtrh_destruct(&ptrh);
    sc_automationPacketPtrh_destruct(&rptrh);
    sm_status_returnError(status, sc_status_failure);    
  }

  status = sc_automationPacketPtrh_destruct(&ptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rptrh);
  sm_status_returnIfFailed(status);  
  
  sm_status_return(status);
}
