#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <unistd.h>

#include <si_sha/insteon.h>
#include <si_sha/autopkt.h>


#define MF_BCAST 0x80
#define MF_NAK   0x80
#define MF_GROUP 0x40
#define MF_ACK   0x20
#define MF_EXTM  0x10

#define MF_HOPSLEFT(v) ((v & 03) << 2)
#define MF_MAXHOPS(v) (v & 03)


static sc_status sc_insteonBridge_handleResponse(sc_insteonBridge* pb,
                                                 const s_uint8* res);

sc_status sc_insteonBridge_serviceDevice(sc_insteonBridge* pb);
sc_status sc_insteonBridge_processRequest(sc_insteonBridge* pb,
                                          sc_automationRequest*  req,
                                          sc_automationResponse* res);

sc_status sc_insteonBridge_fireEvent(sc_insteonBridge* pb,
                                     sc_automationPacket* pkt);


sc_status sc_insteonBridge_construct(sc_insteonBridge* pb,
                                     const char* devName)
{
  static const sc_automationBridge_vtable vtbl = {
    (sc_automationBridge_serviceDeviceF)sc_insteonBridge_serviceDevice,
    (sc_automationBridge_processRequestF)sc_insteonBridge_processRequest,
    (sc_automationBridge_fireEventF)sc_insteonBridge_fireEvent
  };
  sc_status status = sc_status_success;
  sc_insteonBridgeVersion bv;
  char addr[12] = {0};
      
  /* the insteon PLM bridge uses a fixed baud of 19200 */
  status = sc_serialPort_construct(&pb->port_, devName,
                                   se_baudRate_19200);
  sm_status_returnIfFailed(status);

  pb->address_     = 0;
  pb->category_    = 0;
  pb->subcategory_ = 0;

  status = sc_insteonBridge_getVersion(pb, &bv);
  if(sm_status_failed(status)) 
    sc_serialPort_destruct(&pb->port_);
  sm_status_returnIfFailed(status);

  pb->address_     = bv.address_;
  pb->category_    = bv.category_;
  pb->subcategory_ = bv.subcategory_;

  sprintf(addr, "%0X-%0X-%0X", (int)(bv.address_ >> 16),
                               (int)((bv.address_ >> 8) & 0xFF),
                               (int)(bv.address_ & 0xFF));
  
  status = sc_automationBridgeBase_construct(&pb->base_, 24, 16, name,
                                             "03-15", addr, "Insteon",
                                             &vtbl, pb);
  sm_status_returnIfFailed(status);

  pb->base_.fdDev_ = pb->port_.fd_;
  
  sm_status_return(status);
}

sc_status sc_insteonBridge_destruct(sc_insteonBridge* pb)
{
  sc_status status = sc_status_success;

  status = sc_serialPort_destruct(&pb->port_);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}



sc_status sc_insteonBridge_fireEvent(sc_insteonBridge* pb,
                                     sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  sm_status_return(status);    
}
