#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <si_std/string.h>
#include <si_sha/zigbrg.h>
#include <si_sha/instbrg.h>



/* The goal of bridgetool is to:
 * insert into 'bridges' and 'devices' device @ devicePath with
 * identifier 'instanceName'
 *
 */

sc_status insteonBridgeTool(const TCHAR* basePath,
                            const TCHAR* instanceName,
                            const TCHAR* devicePath)
{
  sc_status status = sc_status_success;
  sc_insteonBridge ibrg;
  sc_insteonBridgeVersion bv;
  char sbuff[32];

  status = sc_insteonBridge_construct(&ibrg,
                                      instanceName,
                                      devicePath,
                                      basePath);
  sm_status_returnIfFailed(status);

  status = sc_insteonBridge_getVersion(&ibrg, &bv);
  sm_status_returnIfFailed(status);

  /* print out the bridge address & type at dev path */  
  aprint(sbuff, bv.address_);
  printf(">> %s, ", sbuff);
  tprint(sbuff, bv.category_, bv.subcategory_);  
  printf("%s\n", sbuff);

  status = sc_insteonBridge_destruct(&ibrg);
  sm_status_returnIfFailed(status);  
  
  sm_status_return(status);
}

sc_status zigbeeBridgeTool(const TCHAR* basePath,
                           const TCHAR* instanceName,
                           const TCHAR* devicePath)
{
  sc_status status = sc_status_success;
  sc_zigbeeBridge zbrg;
  char sbuff[48];

  status = sc_zigbeeBridge_construct(&zbrg,
                                     instanceName,
                                     devicePath,
                                     basePath);
  sm_status_returnIfFailed(status);

  zb_aprint(sbuff, zbrg.address_);
  printf(">> %s, ", sbuff);  
  zb_tprint(sbuff, zbrg.type_);
  printf("%s\n", sbuff);

  status = sc_zigbeeBridge_destruct(&zbrg);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}


/*
 * TODO: fix the auto-naming to update the database
 * 
 * This program (or in cooperation with lase-bridge-cfg.pl) should:
 * 1) Let the user specify the serial port the bridge is attached to
 * 2) Let the user specify the type and identifier for the bridge.
 * 3) The program should then connect to the bridge, get address and
 *    network device type, and fill in the devices and bridges tables.
 *
 * Must consider:
 *
 * 1) whether the device already exists in the device table.
 *    if so, does the identifier match the given one?
 * 2) whether the device already exists in the bridges table.
 *    if so, does the device path match the given one?
 * 3) Whether the identifier is valid and not already in use by another
 *    device.
 * 
 */


int isValidIdentifier(const char* id)
{  
  char ch;

  if(id == NULL)
    return 0;
  
  while(ch = *id) {
    if( isalnum(ch) || (ch == '_') )
      id++;
    else
      break;
  }
  
  return (ch == '\0');
}

/*
int isIdentifierAvailableFor(const char* ident, const char* addr)
{
  sqlite3 db;

  sqlite3_open();

  sqlite3_exec();
  
  sqlite3_close();
}
*/

/* could get fancier than this, but no time */
const char* getRootDir(const char* name) {
  if(strstr(name, "lucs-lase-brgtool") != 0)
    return ".";
  else return "/var/lib/lase";
}


int main(int argc, char* argv[])
{
  const char* basePath = getRootDir(argv[0]);
  const char *devPath, *netType;
  char pbuff[1024];
  
  if(argc != 3)
    return 99;

  netType = argv[1];
  devPath = argv[2];

  if(_stricmp(netType, "insteon") == 0) {
    insteonBridgeTool(basePath, "foo.sock", devPath);
  } else if(_stricmp(netType, "zigbee") == 0) {
    zigbeeBridgeTool(basePath, "foo.sock", devPath);
  } else {
    fprintf(stderr, "unknown network type: %s\n", netType);
    return 99;
  }

  sprintf(pbuff, basePath);
  strcat(pbuff, "/foo.sock");  
  unlink(pbuff);
    
  return 0;
}
