#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <si_sha/autopkt.h>
#include <si_sha/autobrg.h>
#include <si_sha/autodev.h>
#include <jsapi.h>


static void js_finalizePacket(JSContext* jscx, JSObject* jso);

static JSClass autoPktClass = {
  "AutomationPacket", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  js_finalizePacket,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

static void js_finalizePacket(JSContext* jscx, JSObject* jso)
{
  sc_automationPacket* pkt;
  /* release the reference held by JSObject */
  pkt = (sc_automationPacket*)JS_GetPrivate(jscx,jso);
  if(pkt)
    pkt->_rep.vtbl->relRef(&pkt->_rep);
}


static JSClass autoBrgClass = {
  "AutomationBridge", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub, /* will need to use this */
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass autoDevClass = {
  "AutomationDevice", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

JSObject *g_devProto = NULL, *g_brgProto = NULL, *g_pktProto;

JSClass* js_getPacketClass()
{
  return &autoPktClass;
}

#define sm_jsPacketPush(N,T)\
static JSBool js_push##N(JSContext *jscx, JSObject *obj,\
                  uintN argc, jsval *argv, jsval *rval)\
{  \
  sc_status status=  sc_status_success;\
  sc_automationPacket* pkt = NULL;\
  T val;\
  if(argc != 1)\
    return JS_FALSE;\
  val = (T)JSVAL_TO_INT(argv[0]);\
  pkt = (sc_automationPacket*) JS_GetInstancePrivate(jscx, obj,\
                                            &autoPktClass, NULL);  \
  status = sc_automationPacket_push##N(pkt, val);\
  if(sm_status_failed(status))\
    return JS_FALSE;\
  *rval = INT_TO_JSVAL((int)val);\
  return JS_TRUE;\
}

#define sm_jsPacketPop(N,T)\
static JSBool js_pop##N(JSContext *jscx, JSObject *obj,\
                      uintN argc, jsval *argv, jsval *rval)\
    {  \
      sc_status status=  sc_status_success;\
      sc_automationPacket* pkt = NULL;\
      T val;\
      if(argc != 0)\
        return JS_FALSE;\
      pkt = (sc_automationPacket*) JS_GetInstancePrivate(jscx, obj,\
                                             &autoPktClass, NULL);  \
      status = sc_automationPacket_pop##N(pkt, &val);\
      if(sm_status_failed(status))\
        return JS_FALSE;\
      *rval = INT_TO_JSVAL((int)val);\
      return JS_TRUE;\
    }

sm_jsPacketPush(UInt8,s_uint8);
sm_jsPacketPush(Int8,s_int8);
sm_jsPacketPush(UInt16,s_uint16);
sm_jsPacketPush(Int16,s_int16);
sm_jsPacketPush(UInt32,s_uint32);
sm_jsPacketPush(Int32,s_int32);
sm_jsPacketPush(UInt64,s_uint64);
sm_jsPacketPush(Int64,s_int64);
sm_jsPacketPop(UInt8,s_uint8);
sm_jsPacketPop(Int8,s_int8);
sm_jsPacketPop(UInt16,s_uint16);
sm_jsPacketPop(Int16,s_int16);
sm_jsPacketPop(UInt32,s_uint32);
sm_jsPacketPop(Int32,s_int32);
sm_jsPacketPop(UInt64,s_uint64);
sm_jsPacketPop(Int64,s_int64);

static JSBool js_tell(JSContext *jscx, JSObject *obj,
                      uintN argc, jsval *argv, jsval *rval)
{  
  sc_status status = sc_status_success;
  sc_automationPacket* pkt = NULL;
  s_uint32 val;
  
  if(argc != 0)
    return JS_FALSE;
  
  pkt = (sc_automationPacket*) JS_GetInstancePrivate(jscx, obj,
                                            &autoPktClass, NULL); 
  status = sc_automationPacket_tell(pkt, &val);
  if(sm_status_failed(status))
    return JS_FALSE;
  *rval = INT_TO_JSVAL((int)val);
  return JS_TRUE;
}

static JSBool js_seek(JSContext *jscx, JSObject *obj,
                      uintN argc, jsval *argv, jsval *rval)
{  
  sc_status status=  sc_status_success;
  sc_automationPacket* pkt = NULL;
  s_uint32 pos;
  if(argc != 1)
    return JS_FALSE;
  
  pos = (s_uint32)JSVAL_TO_INT(argv[0]);
  pkt = (sc_automationPacket*) JS_GetInstancePrivate(jscx, obj,
                                               &autoPktClass, NULL);
  status = sc_automationPacket_setPos(pkt, pos);
  if(sm_status_failed(status))
    return JS_FALSE;
  return JS_TRUE;
}

/* this one can be used for debugging if desired
static JSBool js_pushInt32(JSContext *jscx, JSObject *obj,
                         uintN argc, jsval *argv, jsval *rval)
{  
  sc_status status=  sc_status_success;
  sc_automationPacket* pkt = NULL;
  s_int32 val;
  if(argc != 1)
    return JS_FALSE;
  val = (s_int32)JSVAL_TO_INT(argv[0]);
  pkt = (sc_automationPacket*) JS_GetInstancePrivate(jscx, obj,
                                               &autoPktClass, NULL);
  status = sc_automationPacket_pushInt32(pkt, val);
  if(sm_status_failed(status))
    return JS_FALSE;
  return JS_TRUE;
}
*/

static sc_automationPacketPool pktPool;

JSBool js_autoPacket_construct(JSContext *jscx, JSObject *obj,
                               uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  JSObject* jso = NULL;
  JSBool    rv;
  sc_automationPacketPtrh ptrh;
  sc_automationPacket* pkt = NULL;

  /* get a C packet object from pool */
  status = sc_automationPacketPtrh_construct(&ptrh, NULL);
  if(sm_status_failed(status)) { return JS_FALSE; }
  
  status = sc_automationPacketPool_allocate(&pktPool, &ptrh);
  if(sm_status_failed(status)) { return JS_FALSE; }

  pkt = sc_automationPacketPtrh_get(&ptrh);
  
  /* create js wrapper */
  jso = JS_NewObject(jscx, &autoPktClass, NULL, NULL);
  if(jso == NULL) {
    perror("Unable to create JavaScript object: ");
    return JS_FALSE;
  }
  
  /* attach it to C object */
  rv = JS_SetPrivate(jscx, jso, pkt);
  if(!rv) {
    perror("JS_SetPrivate failed.");
    return JS_FALSE;
  }

  *rval = OBJECT_TO_JSVAL(jso);

  /* note: do not destruct local pointer-handle here. Will be done when
   * the object gets GC'd, in js_finalizePacket. */
  
  return JS_TRUE;
}

static 
   JSBool js_automationDevice_getName(JSContext *cx, JSObject *obj,
                                      jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = NULL;
  
  pd = (sc_automationDevice*)JS_GetPrivate(cx,obj);
  if(pd) {
    *vp = STRING_TO_JSVAL(JS_NewStringCopyZ(cx,pd->name_));
  }
  return JS_TRUE;
}
static 
   JSBool js_automationDevice_getAddress(JSContext *cx, JSObject *obj,
                                         jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = NULL;

  pd = (sc_automationDevice*)JS_GetPrivate(cx, obj);
  if(pd) {
    *vp = STRING_TO_JSVAL(JS_NewStringCopyZ(cx,pd->address_));
  }  
  return JS_TRUE;
}
static 
   JSBool js_automationDevice_getDevType(JSContext *cx, JSObject *obj,
                                         jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = NULL;

  pd = (sc_automationDevice*)JS_GetPrivate(cx, obj);
  if(pd) {
    *vp = STRING_TO_JSVAL(JS_NewStringCopyZ(cx,pd->type_));
  }  
  return JS_TRUE;
}
static 
   JSBool js_automationDevice_getNetType(JSContext *cx, JSObject *obj,
                                         jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = NULL;

  pd = (sc_automationDevice*)JS_GetPrivate(cx, obj);
  if(pd) {
    *vp = STRING_TO_JSVAL(JS_NewStringCopyZ(cx,pd->netType_));
  }  
  return JS_TRUE;
}

static JSBool js_rename(JSContext *cx, JSObject *obj,
                        uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = NULL;
  JSString* str;
  char* bytes = NULL;

  pd = (sc_automationDevice*)JS_GetPrivate(cx,obj);
  if(pd) {
    str = JS_ValueToString(cx, argv[0]);
    bytes = JS_EncodeString(cx, str);
    if(pd->name_) free((void*)pd->name_);
    pd->name_ = strdup(bytes);
    JS_free(cx, bytes);    
  } else return JS_FALSE;
  return JS_TRUE;
}

static sc_automationPacket* pktFromJsval(JSContext *cx, jsval ov)
{
  sc_automationPacket* pk = NULL;
  JSObject* po = JSVAL_TO_OBJECT(ov);
  pk = (sc_automationPacket*)JS_GetPrivate(cx,po);
  return  pk;
}

static JSBool js_sendMessage(JSContext *cx, JSObject *obj,
                             uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_automationBridgeDevice* pb = NULL;
  pb = (sc_automationBridgeDevice*)JS_GetPrivate(cx,obj);
  if(pb) {
    sc_automationPacketPtrh rsptrh;
    sc_automationPacket* pres = NULL;
    JSObject* jsp = NULL;

    sc_automationPacketPtrh_construct(&rsptrh, NULL);
    
    status = sc_automationBridgeDevice_sendMessage(pb,
                                      pktFromJsval(cx,argv[0]), &rsptrh);
    
    /* make a new packet to return to javascript */
    jsp = JS_NewObject(cx, &autoPktClass, g_pktProto, NULL);
    if(jsp == NULL) {
      JS_ReportError(cx, "OutOfMemory");
      return JS_FALSE;
    }
    
    pres = sc_automationPacketPtrh_get(&rsptrh);

    JS_SetPrivate(cx, jsp, pres);
    *rval = OBJECT_TO_JSVAL(jsp);
    
    /* NOTE: do not destruct ptrh here */
  }    
  return JS_TRUE;
}
static JSBool js_newPacket(JSContext *cx, JSObject *jsb,
                           uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_automationBridgeDevice* pb = NULL;
  
  pb = (sc_automationBridgeDevice*)JS_GetPrivate(cx, jsb);
  if(pb) {
    sc_automationPacketPtrh ptrh;
    sc_automationPacket* pkt = NULL;
    JSObject* jsp = NULL;

    sc_automationPacketPtrh_construct(&ptrh, NULL);

    /* ask bridge to allocate packet */
    status = sc_automationBridgeDevice_newPacket(pb, &ptrh);
    if(sm_status_failed(status)) {
      JS_ReportError(cx, "OutOfMemory");
      return JS_FALSE;
    }
    
    /* make a new packet to return to javascript */
    jsp = JS_NewObject(cx, &autoPktClass, g_pktProto, NULL);
    if(jsp == NULL) {
      JS_ReportError(cx, "OutOfMemory");
      return JS_FALSE;
    }
    
    /* handle possible memory errors here */    
    pkt = sc_automationPacketPtrh_get(&ptrh);
    JS_SetPrivate(cx, jsp, pkt);
    *rval = OBJECT_TO_JSVAL(jsp);
    
    /* NOTE: do not destruct ptrh here */
  }    
  return JS_TRUE;
}

static JSBool js_discover(JSContext *cx, JSObject *jsb,
                          uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_automationBridgeDevice* pb = NULL;
  JSObject *jsgbl, *jsd;
  const TCHAR* brgName = NULL;
  int nidx = 0;
  
  pb = (sc_automationBridgeDevice*)JS_GetPrivate(cx, jsb);
  if(!pb) {
    return JS_FALSE; /* error? */
  }
    
  status = sc_automationBridgeDevice_discover(pb);
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE; /* error? */
  }
      
  *rval = BOOLEAN_TO_JSVAL(JS_TRUE);
  return JS_TRUE;
}


static JSFunctionSpec pktfs[] = {
  { "pushUInt8",  js_pushUInt8,  1, 0, 0},
  { "pushUInt16", js_pushUInt16, 1, 0, 0},
  { "pushUInt32", js_pushUInt32, 1, 0, 0},
  { "pushUInt64", js_pushUInt64, 1, 0, 0},
  { "pushInt8",   js_pushInt8,   1, 0, 0},
  { "pushInt16",  js_pushInt16,  1, 0, 0},
  { "pushInt32",  js_pushInt32,  1, 0, 0},
  { "pushInt64",  js_pushInt64,  1, 0, 0},
  { "popUInt8",   js_popUInt8,   0, 0, 0},
  { "popUInt16",  js_popUInt16,  0, 0, 0},
  { "popUInt32",  js_popUInt32,  0, 0, 0},
  { "popUInt64",  js_popUInt64,  0, 0, 0},
  { "popInt8",    js_popInt8,    0, 0, 0},
  { "popInt16",   js_popInt16,   0, 0, 0},
  { "popInt32",   js_popInt32,   0, 0, 0},
  { "popInt64",   js_popInt64,   0, 0, 0},
  { "tell",       js_tell,       0, 0, 0},
  { "seek",       js_seek,       1, 0, 0},
  JS_FS_END 
};

static JSPropertySpec devps[] = {
  {"networkType", 0, JSPROP_READONLY | JSPROP_ENUMERATE, js_automationDevice_getNetType, NULL},
  {"address",     0, JSPROP_READONLY | JSPROP_ENUMERATE, js_automationDevice_getAddress, NULL},
  {"deviceType",  0, JSPROP_READONLY | JSPROP_ENUMERATE, js_automationDevice_getDevType, NULL},
  {"name",        0, JSPROP_READONLY | JSPROP_ENUMERATE, js_automationDevice_getName, NULL},
  {0}
};
static JSFunctionSpec devfs[] = {
  { "_rename",  js_rename,  1, 0, 0},
  JS_FS_END
};

static JSFunctionSpec brgfs[] = {
  { "sendMessage",  js_sendMessage,  1, 0, 0},
  { "newPacket",    js_newPacket, 0, 0, 0},
  { "discover",     js_discover, 0, 0, 0},
  JS_FS_END 
};


/* TODO: clean this up later */
extern sc_status js_installInsteonClasses(JSContext* ctxt);
extern sc_status js_installZigbeeClasses(JSContext* ctxt);


sc_status js_installAutomationClasses(JSContext* ctxt)
{
  sc_status status = sc_status_success;  
  JSObject* nullProto = NULL;  
  JSObject* global = JS_GetGlobalObject(ctxt);
  
  g_pktProto = JS_InitClass(ctxt, global, nullProto, &autoPktClass,
                          js_autoPacket_construct, 0,
                          NULL, pktfs, NULL, NULL);

  g_devProto = JS_InitClass(ctxt, global, nullProto, &autoDevClass,
                          NULL, 0, devps, devfs, NULL, NULL);
  
  g_brgProto = JS_InitClass(ctxt, global, g_devProto, &autoBrgClass,
                          NULL, 0, NULL, brgfs, NULL, NULL);
  

  /* this will suffice for now but better to have bridge-specific pools */
  status = sc_automationPacketPool_construct(&pktPool, 256, 32, 4096);
  sm_status_returnIfFailed(status);

  status = js_installInsteonClasses(ctxt);
  sm_status_returnIfFailed(status);
  status = js_installZigbeeClasses(ctxt);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}
