#include <stdlib.h>
#include <string.h>

#include <si_sha/autopkt.h>
#include <si_sha/instbrg.h>
#include <si_sha/instbdev.h>
#include <jsapi.h>

static JSBool js_sendStdMessage(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval);

static JSBool js_sendStdWaitAck(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval);
static JSBool js_resetDeviceDB(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval);

extern JSClass* js_getPacketClass();



static JSFunctionSpec objfs[] = {
  { "sendStdMessage",   js_sendStdMessage,   3, 0, 0},
  { "sendStdWaitAck",   js_sendStdWaitAck,   3, 0, 0},
  { "clearDeviceNet",   js_resetDeviceDB,    0, 0, 0},
  JS_FS_END
};

static JSClass instBrgClass = {
  "InsteonBridge", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};
   
static JSBool js_sendStdMessage(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_insteonBridgeDevice* pb = NULL;
  
  pb = (sc_insteonBridgeDevice*)JS_GetInstancePrivate(cx, jsb,
                                                &instBrgClass,argv);
  if(!pb) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;     /* error? */
  }

  status = sc_insteonBridgeDevice_sendStdMessage(pb,
                             JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
                             (s_uint8)JSVAL_TO_INT(argv[1]),
                             (s_uint8)JSVAL_TO_INT(argv[2]));
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;
  }

  *rval = BOOLEAN_TO_JSVAL(JS_TRUE);  
  return JS_TRUE;
}

static JSBool js_sendStdWaitAck(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_insteonBridgeDevice* pb = NULL;
  sc_automationPacketPtrh rptrh;
  sc_automationPacket* pkt = NULL;
  JSObject* jspo;
  JSBool rv;

  status = sc_automationPacketPtrh_construct(&rptrh, NULL);
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;
  }  
  
  pb = (sc_insteonBridgeDevice*)JS_GetInstancePrivate(cx, jsb,
                                              &instBrgClass, argv);
  if(!pb) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;     /* error? */
  }

  status = sc_insteonBridgeDevice_sendStdWaitAck(pb,
                             JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
                             (s_uint8)JSVAL_TO_INT(argv[1]),
                             (s_uint8)JSVAL_TO_INT(argv[2]),
                             &rptrh);
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;
  }

  /* create js wrapper for packet */
  jspo = JS_NewObject(cx, js_getPacketClass(), NULL, NULL);
  if(jspo == NULL) 
    return JS_FALSE;
  pkt = sc_automationPacketPtrh_get(&rptrh);
  
  /* attach it to C object */
  rv = JS_SetPrivate(cx, jspo, pkt);
  if(!rv) 
    return JS_FALSE;  
  *rval = OBJECT_TO_JSVAL(jspo);
  
  return JS_TRUE;
}

static JSBool js_resetDeviceDB(JSContext *cx, JSObject *jsb,
                               uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;  
  sc_insteonBridgeDevice* pb = NULL;
  
  pb = (sc_insteonBridgeDevice*)JS_GetInstancePrivate(cx, jsb,
                                                      &instBrgClass, argv);
  if(!pb) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;     /* error? */
  }
           
  status = sc_insteonBridgeDevice_resetDeviceDB(pb);
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;
  }
  *rval = BOOLEAN_TO_JSVAL(JS_TRUE);
  return JS_TRUE;    
}

extern JSObject* g_brgProto;

JSObject* g_instBrgProto = NULL;

sc_status js_installInsteonClasses(JSContext* ctxt)
{  
  sc_status status = sc_status_success;
  
  g_instBrgProto = JS_InitClass(ctxt, JS_GetGlobalObject(ctxt), g_brgProto,
                                &instBrgClass,
                                NULL, 0,
                                NULL, objfs, NULL, NULL);
  
  sm_status_return(status);
}

sc_status js_createInsteonBridgeDevice(JSContext* ctxt,
                                       sc_automationDeviceMgr* pdm,
                                       const TCHAR* name,
                                       const TCHAR* devName,
                                       const TCHAR* bridgePath)
{
  sc_status status= sc_status_success;
  sc_insteonBridge* ibrg;
  sc_insteonBridgeDevice* idev;
  JSObject* jso = NULL;

  ibrg = malloc(sizeof(*ibrg));
  if(ibrg == NULL) {
    sm_status_returnError(status, sc_err_outOfMemory);
  }
  idev = malloc(sizeof(*idev));
  if(ibrg == NULL) {
    free(idev);
    sm_status_returnError(status, sc_err_outOfMemory);
  }
  
  status = sc_insteonBridge_construct(ibrg, name, devName, bridgePath);
  sm_status_returnIfFailed(status);
  status = sc_automationBridgeBase_start(&ibrg->base_);
  sm_status_returnIfFailed(status);
  
  status = sc_insteonBridgeDevice_construct(idev, pdm, name, bridgePath);
  sm_status_returnIfFailed(status);

  jso = JS_DefineObject(ctxt, JS_GetGlobalObject(ctxt), name,
                        &instBrgClass, g_instBrgProto,
                        JSPROP_READONLY | JSPROP_ENUMERATE);
  JS_SetPrivate(ctxt, jso, idev);
  
  sm_status_return(status);
}
