#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <si_sha/autopkt.h>
#include <si_sha/zigbrg.h>
#include <si_sha/zigbdev.h>
#include <jsapi.h>

static JSBool js_zbSendMessage(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval);

extern JSClass* js_getPacketClass();

static JSFunctionSpec objfs[] = {
  { "sendMessage",   js_zbSendMessage,  3, 0, 0},
  JS_FS_END
};

static JSClass zigbeeBrgClass = {
  "ZigBeeBridge", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSBool js_zbSendMessage(JSContext *cx, JSObject *jsb,
                                uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  sc_zigbeeBridgeDevice* pzb = NULL;

  pzb = (sc_zigbeeBridgeDevice*)JS_GetInstancePrivate(cx, jsb,
                                        &zigbeeBrgClass,argv);
  if(!pzb) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;     /* error? */
  }

  /*
  status = sc_zigbeeBridgeDevice_sendMessage(pzb,
                         JS_GetStringBytes(JSVAL_TO_STRING(argv[0])),
    (s_uint8)JSVAL_TO_INT(argv[1]),
    (s_uint8)JSVAL_TO_INT(argv[2]));
  if(sm_status_failed(status)) {
    *rval = BOOLEAN_TO_JSVAL(JS_FALSE);
    return JS_FALSE;
  }
  */
  
  *rval = BOOLEAN_TO_JSVAL(JS_TRUE);  
  return JS_TRUE;
}


extern JSObject* g_brgProto;

JSObject* g_zigbBrgProto = NULL;

sc_status js_installZigbeeClasses(JSContext* ctxt)
{  
  sc_status status = sc_status_success;

  g_zigbBrgProto = JS_InitClass(ctxt, JS_GetGlobalObject(ctxt), g_brgProto,
                                &zigbeeBrgClass,
                                NULL, 0,
                                NULL, objfs, NULL, NULL);

  sm_status_return(status);
}

const char* argByName(const char* name, const char* args)
{
  int nameLen;
  const char* str = NULL;

  nameLen = strlen(name);
  str = strstr(args, name);

  while(str != NULL) {
    const char* ch = str + nameLen;
    if(*ch == '=')
      return str_advw(ch+1);
    else if(iswhitespace(*ch)) {
      ch = str_advw(ch);
      if(*ch == '=') 
        return str_advw(ch+1);
    }
    str = strstr(ch, name); /* could be a partial match of longer name */
  }
  return NULL;
}

sc_status js_createZigbeeBridgeDevice(JSContext* ctxt,
                                      sc_automationDeviceMgr* pdm,
                                      const TCHAR* name,
                                      const TCHAR* devName,
                                      const TCHAR* bridgePath,
                                      const TCHAR* startupArgs)
{
  sc_status status= sc_status_success;
  sc_zigbeeBridge* zbrg;
  sc_zigbeeBridgeDevice* zdev;
  JSObject* jso = NULL;
  s_uint8 channel = 11;
  s_uint16 panId = 0xbabe;
  const char* tmp;

  tmp = argByName("channel", startupArgs);
  if(tmp != NULL)
    channel = (s_uint8) atoi(tmp);
  tmp = argByName("panId", startupArgs);
  if(tmp != NULL)
    panId = (s_uint16) strtoul(tmp, NULL, 0);
  
  zbrg = malloc(sizeof(*zbrg));
  if(zbrg == NULL) {
    sm_status_returnError(status, sc_err_outOfMemory);
  }
  zdev = malloc(sizeof(*zdev));
  if(zdev == NULL) {
    free(zbrg);
    sm_status_returnError(status, sc_err_outOfMemory);
  }

  status = sc_zigbeeBridge_construct(zbrg, name, devName, bridgePath);
  sm_status_returnIfFailed(status);
  status = sc_automationBridgeBase_start(&zbrg->base_);
  sm_status_returnIfFailed(status);

  status = sc_zigbeeBridgeDevice_construct(zdev, pdm, name, bridgePath,
                                           channel, panId);
  sm_status_returnIfFailed(status);

  jso = JS_DefineObject(ctxt, JS_GetGlobalObject(ctxt), name,
                        &zigbeeBrgClass, g_zigbBrgProto,
                        JSPROP_READONLY | JSPROP_ENUMERATE);
  JS_SetPrivate(ctxt, jso, zdev);

  sm_status_return(status);
}
