#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <si_std/string.h>
#include <si_sha/lase.h>

#include <unistd.h>
#include <poll.h>


/* TODO: put these somewhere better */
sc_status js_installAutomationClasses(JSContext* ctxt,
                                      JSObject* pglb);

sc_status js_createInsteonBridgeDevice(JSContext* ctxt,
                                       sc_automationDeviceMgr* pdm,
                                       const TCHAR* name,
                                       const TCHAR* devName,
                                       const TCHAR* bridgePath);

sc_status js_createZigbeeBridgeDevice(JSContext* ctxt,
                                      sc_automationDeviceMgr* pdm,
                                      const TCHAR* name,
                                      const TCHAR* devName,
                                      const TCHAR* bridgePath,
                                      const TCHAR* startupArgs);
                                      

static void* svcEntry(void* arg);
extern JSClass* js_getPacketClass();
sc_status sc_lase_loadBridges(sc_lase* pl);

static JSClass globclass = {
  "global", JSCLASS_GLOBAL_FLAGS | JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

/* The error reporter callback. Err, until I know what to do here */
void reportError(JSContext *cx, const char *message, JSErrorReport *report)
{   
  fprintf(stderr, "%s:%u:%s\n",
          report->filename ? report->filename : "<no filename>",
          (unsigned int) report->lineno,
          message);
}

/* function below lifted from SpiderMonkey shell */
static JSBool
   js_print(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  uintN i;
  JSString *str;
  char *bytes;

  /* convert all arguments to string and print */
  for (i = 0; i < argc; i++) {
    str = JS_ValueToString(cx, argv[i]);
    if (!str)
      return JS_FALSE;
    bytes = JS_EncodeString(cx, str);
    if (!bytes)
      return JS_FALSE;
    fprintf(stdout, "%s%s", i ? " " : "", bytes);
    JS_free(cx, bytes);
  }

  fputc('\n', stdout);
  fflush(stdout);

  return JS_TRUE;
}
static JSBool
   js_archiveNode(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  JSObject* glob = JS_GetGlobalObject(cx);
  sc_lase* pl = NULL;
  JSString *str;
  char* bytes;
  sc_string ident, archive;

  if(argc != 2)
    return JS_FALSE;
  
  pl = (sc_lase*)JS_GetInstancePrivate(cx, glob, &globclass, NULL);

  sc_string_construct(&ident);
  sc_string_construct(&archive);
  
  str = JS_ValueToString(cx, argv[0]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&ident, bytes);
  JS_free(cx, bytes);

  str = JS_ValueToString(cx, argv[1]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&archive, bytes);
  JS_free(cx, bytes);

  /* device manager resolves name and stores link */
  status = sc_automationDeviceMgr_saveArchive(&pl->dmgr, ident.str_, 
                                              archive.str_);
  
  sc_string_destruct(&ident);
  sc_string_destruct(&archive);
  
  if(sm_status_succeeded(status))
    return JS_TRUE;
  else
    return JS_FALSE;
}


static JSBool getDeviceName(JSContext *cx, JSObject* o, sc_string* nstr)
{ /* get device name by reading name property */
  JSBool rv = JS_TRUE;
  JSString *str;
  char* bytes;
  jsval jsv;
  
  rv = JS_GetProperty(cx, o, "name", &jsv);
  if(!rv)
    return rv;
  if(JSVAL_IS_VOID(jsv))
    return JS_FALSE;
  
  str   = JS_ValueToString(cx, jsv);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(nstr, bytes);
  JS_free(cx, bytes);
  
  return JS_TRUE;
}

static JSBool
   js_clearLinks(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  JSObject *glob = JS_GetGlobalObject(cx), *jsro = NULL;
  sc_lase* pl = NULL;
  JSString *str;
  char* bytes;
  sc_string srcIdent;
  JSBool rv = JS_TRUE;

  if(argc != 1) 
    return JS_FALSE;

  /* get lase instance */
  pl = (sc_lase*) JS_GetInstancePrivate(cx, glob, &globclass, NULL);

  /* argument is the instance to rename */
  jsro = JSVAL_TO_OBJECT(argv[0]);
  if(jsro == NULL)
    return JS_FALSE;

  sc_string_construct(&srcIdent);
  
  rv = getDeviceName(cx, jsro, &srcIdent);
  if(!rv) 
    return rv;
  
  /* clear links for this object */
  status = sc_automationDeviceMgr_clearLinks(&pl->dmgr, srcIdent.str_);
  if(sm_status_failed(status)) {
    sc_string_destruct(&srcIdent);
    return JS_FALSE;
  }
  sc_string_destruct(&srcIdent);

  return JS_TRUE;
}

static JSBool
   js_archiveLink(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{
  sc_status status = sc_status_success;
  JSObject* glob = JS_GetGlobalObject(cx);
  sc_lase* pl = NULL;
  JSString *str;
  char* bytes;
  sc_string srcIdent, dstIdent, srcEvent, dstMethd;

  if(argc != 4)
    return JS_FALSE;

  pl = (sc_lase*)JS_GetInstancePrivate(cx, glob, &globclass, NULL);

  sc_string_construct(&srcIdent);
  sc_string_construct(&dstIdent);
  sc_string_construct(&srcEvent);
  sc_string_construct(&dstMethd);

  str = JS_ValueToString(cx, argv[0]); 
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&srcIdent, bytes); /* copy to string */
  JS_free(cx, bytes);

  str = JS_ValueToString(cx, argv[1]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&srcEvent, bytes); /* copy to string */
  JS_free(cx, bytes); /* free JS string data */

  str = JS_ValueToString(cx, argv[2]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&dstIdent, bytes); /* copy to string */
  JS_free(cx, bytes); /* free JS string data */

  str = JS_ValueToString(cx, argv[3]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&dstMethd, bytes); /* copy to string */
  JS_free(cx, bytes); /* free JS string data */
  
  /* Device manager resolves the names and stores the link */
  status = sc_automationDeviceMgr_saveLink(&pl->dmgr,
                                           srcIdent.str_,
                                           srcEvent.str_,
                                           dstIdent.str_,
                                           dstMethd.str_);

  sc_string_destruct(&srcIdent);
  sc_string_destruct(&dstIdent);
  sc_string_destruct(&srcEvent);
  sc_string_destruct(&dstMethd);

  if(sm_status_succeeded(status))
    return JS_TRUE;
  else
    return JS_FALSE;
}

static JSBool js_rename(JSContext *cx, JSObject *obj, uintN argc, jsval *argv, jsval *rval)
{ /* rename object */
  sc_status status = sc_status_success;
  JSObject* glob = JS_GetGlobalObject(cx), *jsro = NULL;
  sc_lase* pl = NULL;
  JSString *str;
  char* bytes;
  sc_string ident, nident;
  JSBool rv;
  
  if(argc != 2)
    return JS_FALSE;

  pl = (sc_lase*)JS_GetInstancePrivate(cx, glob, &globclass, NULL);

  sc_string_construct(&ident);
  sc_string_construct(&nident);

  // first argument is the instance to rename
  jsro = JSVAL_TO_OBJECT(argv[0]);
  if(jsro == NULL)
    return JS_FALSE;

  /* get current name */
  rv = getDeviceName(cx, jsro, &ident);
  if(!rv)
    return rv;
  
  str = JS_ValueToString(cx, argv[1]);
  bytes = JS_EncodeString(cx, str);
  sc_string_assign(&nident, bytes);
  JS_free(cx, bytes);

  /* Device manager handles renaming */
  status = sc_automationDeviceMgr_renameDevice(&pl->dmgr, ident.str_, nident.str_);
  if(sm_status_failed(status)) {
    sc_string_destruct(&ident);    
    sc_string_destruct(&nident);
    return JS_FALSE;
  }
  sc_string_destruct(&ident);
  sc_string_destruct(&nident);
  
  return JS_TRUE;
}

static sc_status loadScript(JSContext* cx, const char* filename)
{
  sc_status status = sc_status_success;
  JSScript *script;
  JSBool ok;
  jsval result;
  JSObject* obj;
  s_uint32 oldopts;

  obj = JS_GetGlobalObject(cx);

  /* save options, set compile & go, and execute */
  oldopts = JS_GetOptions(cx);

  JS_SetOptions(cx, oldopts | JSOPTION_COMPILE_N_GO);
  
  /* load and compile filename to script object */
  script = JS_CompileFile(cx, obj, filename);
  if(!script) {
    sm_status_returnError(status, sc_status_failure);
  }
  /* execute script */
  ok = JS_ExecuteScript(cx, obj, script, &result);

  /* free script & restore options */
  JS_DestroyScript(cx, script);
  JS_SetOptions(cx, oldopts);

  sm_status_return(status);
}

static void* svcEntry(void* arg)
{
  sc_status status = sc_status_success;
  sc_lase* pl = (sc_lase*)arg;

  status = sc_lase_serviceThread(pl);
  assert(status == sc_status_success);
  
  return NULL;
}

sc_status sc_lase_construct(sc_lase* pl, s_uint32 envSize,
                            const TCHAR* basePath)
{
  sc_status status = sc_status_success;
  pthread_mutexattr_t mattr;
  int rv;
  sc_string path;
  
  status = sc_string_constructFrom(&pl->basePath, basePath);
  sm_status_returnIfFailed(status);
  status = sc_scriptRequestPool_construct(&pl->pool, 8, 32);
  sm_status_returnIfFailed(status);
  status = sc_scriptRequestPtrhQueue_construct(&pl->reqs, 8);
  sm_status_returnIfFailed(status);

  pthread_mutexattr_init(&mattr);
  pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE_NP);
  pthread_mutex_init(&pl->mux, &mattr);
  
  /* new runtime with an 8MB memory area */
  pl->jsrt = JS_NewRuntime(8 * 1024 * 1024);
  if(pl->jsrt == NULL) {
    perror("Unable to create JavaScript runtime: ");
    sm_status_returnError(status, sc_status_failure);
  }
  pl->jscx = JS_NewContext(pl->jsrt, 8192);
  if(pl->jscx == NULL) {
    perror("Unable to create JavaScript context: ");
    sm_status_returnError(status, sc_status_failure);
  }

  /* set context options */
  JS_SetOptions(pl->jscx, JSOPTION_VAROBJFIX);
  JS_SetVersion(pl->jscx, JSVERSION_LATEST);
  JS_SetErrorReporter(pl->jscx, reportError);

  pl->glob = JS_NewObject(pl->jscx, &globclass, NULL, NULL);
  if(pl->glob == NULL) {
    perror("Unable to create JavaScript global: ");
    sm_status_returnError(status, sc_status_failure);
  }
  if(!JS_InitStandardClasses(pl->jscx, pl->glob)) {
    perror("JS_InitStandardClasses failed:");
    sm_status_returnError(status, sc_status_failure);
  }

  /* install this as the private pointer in global object */
  rv = JS_SetPrivate(pl->jscx, pl->glob, pl);
  if(!rv) {
    perror("Global JS_SetPrivate failed.");
    return JS_FALSE;
  }  

  /* add print function */
  if (!JS_DefineFunction(pl->jscx, pl->glob, "print",
                         js_print, 0, 0)) {
    perror("JS_DefineFunction failed:");
    sm_status_returnError(status, sc_status_failure);
  }
  /* add the following functions. They don't appear at global scope
   * because they begin with _ */
  if (!JS_DefineFunction(pl->jscx, pl->glob, "_archiveDevice",
                         js_archiveNode, 0, 0)) {
    perror("JS_DefineFunction failed:");
    sm_status_returnError(status, sc_status_failure);
  }
  if (!JS_DefineFunction(pl->jscx, pl->glob, "_archiveLink",
                         js_archiveLink, 0, 0)) {
    perror("JS_DefineFunction failed:");
    sm_status_returnError(status, sc_status_failure);
  }
  if (!JS_DefineFunction(pl->jscx, pl->glob, "_clearLinks",
                         js_clearLinks, 0, 0)) {
    perror("JS_DefineFunction failed:");
    sm_status_returnError(status, sc_status_failure);
  }
  if (!JS_DefineFunction(pl->jscx, pl->glob, "_rename",
                         js_rename, 0, 0)) {
    perror("JS_DefineFunction failed:");
    sm_status_returnError(status, sc_status_failure);
  }
  
  
  status = sc_automationDeviceMgr_construct(&pl->dmgr, pl);
  sm_status_returnIfFailed(status);

  /* init built in automation classes */
  status = js_installAutomationClasses(pl->jscx, pl->glob);
  sm_status_returnIfFailed(status);  

  /* all bridges must be loaded at this point */
  status = sc_lase_loadBridges(pl);
  sm_status_returnIfFailed(status);

  status = sc_string_constructFrom(&path, pl->basePath.str_);
  sm_status_returnIfFailed(status);
  status = sc_string_append(&path, "/js/auto.js");
  sm_status_returnIfFailed(status);  
  /* load main javascript file */
  status = loadScript(pl->jscx, path.str_);
  sm_status_returnIfFailed(status);
  
  status = sc_string_assign(&path, pl->basePath.str_);
  sm_status_returnIfFailed(status);
  status = sc_string_append(&path, "/js/json2.js");
  sm_status_returnIfFailed(status);  
  /* load json support */  
  status = loadScript(pl->jscx, path.str_);
  sm_status_returnIfFailed(status);

  status = sc_string_destruct(&path);
  sm_status_returnIfFailed(status);

  pl->state = se_laseStateStopped;
  
  sm_status_return(status);
}


#include <sqlite3.h>

static const char* fieldByName(const char* field,
                        int argc, char* argv[], char* names[])
{ /* gets the correct value from the argv array using the column */
  /* names from the names array */
  int idx;
  for (idx=0; idx < argc; ++idx) {
    if(_stricmp(names[idx], field) == 0) {
      /* return the value */
      return argv[idx]; 
    }
  }
  return NULL;
}

static int bridgeLoad(void* arg, int argc, char* argv[], char* names[])
{ /* attempt to load the bridge matching this line */
  sc_lase* pl = (sc_lase*)arg;
  const char* ident;
  const char* dtype;
  const char* dpath;
  const char* dargs;

  if(argc < 5) 
    return 0; /* not enough info to load bridge */
  
  ident = fieldByName("identifier", argc, argv, names);
  dtype = fieldByName("type",       argc, argv, names);
  dpath = fieldByName("devPath",    argc, argv, names);
  dargs = fieldByName("args",       argc, argv, names);
  
  if( (ident == NULL) || (dtype == NULL) || (dpath == NULL) )
    return 0; /* not enough info to load bridge */
  
  /* this can be done in a fancier more dynamic way, but I need to
   * finish this thing someday! */
  if(strcmp("01-03-15", dtype) == 0) {
    js_createInsteonBridgeDevice(pl->jscx, &pl->dmgr, ident, dpath,
                                 pl->basePath.str_);
  } else if (strcmp("02-00-01", dtype) == 0) {
    js_createZigbeeBridgeDevice(pl->jscx, &pl->dmgr, ident, dpath,
                                pl->basePath.str_, dargs);
  }
  
  return 0;
}

sc_status sc_lase_loadBridges(sc_lase* pl)
{
  static const char* query = "select b.*, d.type, d.identifier from "
                             "bridges b join devices d on b.deviceId=d.id;";
  sc_status status = sc_status_success;
  int rc;
  sqlite3* db = NULL;
  char *eMsg = NULL;
  sc_string path;

  status = sc_string_constructFrom(&path, pl->basePath.str_);
  sm_status_returnIfFailed(status);
  status = sc_string_append(&path, "/db/lase.db");
  sm_status_returnIfFailed(status);

  /* open database  */
  rc = sqlite3_open(path.str_, &db);
  if(rc != 0) {
    fprintf(stderr, "couldn't open lase.db for startup processing.\n");
    sc_string_destruct(&path);
    sm_status_returnError(status, sc_status_failure);
  }
  
  status = sc_string_destruct(&path);
  sm_status_returnIfFailed(status);

  /* query for bridges in bridge table */
  rc = sqlite3_exec(db, query, bridgeLoad, pl, &eMsg);
  if(rc != 0) {
    fprintf(stderr, "Failed to query bridges: %s\n", eMsg);
    sqlite3_free(eMsg);
  }

  /* close db */
  sqlite3_close(db);
  
  sm_status_return(status);
}



sc_status sc_lase_start(sc_lase* pl)
{
  sc_status status = sc_status_success;
  
  if(pl->state == se_laseStateStopped) {
    int fds[2] = {-1,-1};
    /* pipe for control (start/stop or request available) */
    pipe(fds);
    pl->ctlFdS = fds[0];
    pl->ctlFdC = fds[1];

    pl->state = se_laseStateStarting;
    pthread_create(&pl->svcThread, NULL, svcEntry, pl);
  }
  
  sm_status_return(status);
}

sc_status sc_lase_stop(sc_lase* pl)
{
  sc_status status = sc_status_success;

  if(pl->state == se_laseStateStarted) {
    s_uint8 ch = 0x04;

    pl->state = se_laseStateStopping;
    write(pl->ctlFdC, &ch, 1);
    pthread_join(pl->svcThread, NULL);
  }
  
  sm_status_return(status);
}

sc_status sc_lase_destruct(sc_lase* pl)
{
  sc_status status = sc_status_success;

  status = sc_lase_stop(pl);
  sm_status_returnIfFailed(status);
  
  /* attempt to cleanup everything */
  JS_GC(pl->jscx);

  status = sc_automationDeviceMgr_destruct(&pl->dmgr);
  sm_status_returnIfFailed(status);

  JS_DestroyContext(pl->jscx);
  JS_DestroyRuntime(pl->jsrt);
  JS_ShutDown();
  
  status = sc_scriptRequestPtrhQueue_destruct(&pl->reqs);
  sm_status_returnIfFailed(status);
      
  status = sc_scriptRequestPool_destruct(&pl->pool);
  sm_status_returnIfFailed(status);

  close(pl->ctlFdC);
  close(pl->ctlFdS);
  pthread_mutex_destroy(&pl->mux);
  
  sm_status_return(status);
}

sc_status sc_lase_allocRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh)
{
  return sc_scriptRequestPool_allocate(&pl->pool, ptrh);
}

sc_status sc_lase_submitRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh)
{
  sc_status status = sc_status_success;
  char ch = 0;
  
  pthread_mutex_lock(&pl->mux);
  status = sc_scriptRequestPtrhQueue_pushBack(&pl->reqs, ptrh);
  pthread_mutex_unlock(&pl->mux);
  
  sm_status_returnIfFailed(status); /* check status from locked code */
  
  write(pl->ctlFdC, &ch, 1); /* signal svcThread to execute it */

  sm_status_return(status);
}

JSBool sc_lase_relayMessage(sc_lase* pl, sc_scriptRequest* psr, jsval* prtv)
{
  sc_status status = sc_status_success;
  int rv = 0;
  sc_automationPacketPtrh lptrh;
  jsval arg, obj;
  JSObject *jsro, *jspo;
  
  sc_automationPacketPtrh_construct(&lptrh, &psr->pkt->_rep);
  /* note, ptrh above intentionally not destroyed */
    
  /* create js wrapper */
  jspo = JS_NewObject(pl->jscx, js_getPacketClass(), NULL, NULL);
  if(jspo == NULL) 
    return JS_FALSE;
  
  /* attach it to C object */
  rv = JS_SetPrivate(pl->jscx, jspo, psr->pkt);
  if(!rv) 
    return JS_FALSE;  
  arg = OBJECT_TO_JSVAL(jspo);

  /* get js obj-ref */
  rv = JS_EvaluateScript(pl->jscx, pl->glob, psr->obj, strlen(psr->obj),
                         "script", 0, &obj);
  if(!rv) 
    return JS_FALSE;
  jsro = JSVAL_TO_OBJECT(obj);
  
  rv = JS_CallFunctionName(pl->jscx, jsro, "_relayMessage", 1, &arg, prtv);
  if(!rv) 
    return JS_FALSE;
  
  return JS_TRUE;
}

sc_status sc_lase_executeHere(sc_lase* pl, sc_scriptRequestPtrh* ptrh)
{
  sc_status status = sc_status_success;
  sc_scriptRequest* preq = NULL;
  int rv=0, lnum=0;
  jsval rval;

  preq = sc_scriptRequestPtrh_get(ptrh);
  if(preq == NULL)
    sm_status_return(status);

  if((preq->flags & se_scriptRequest_RelayMessage) != 0) {
    /* relayed message */
    rv = sc_lase_relayMessage(pl, preq, &rval);
  } else {
    /* script request, passed directly to SpiderMonkey API */
    rv = JS_EvaluateScript(pl->jscx, pl->glob,
                           preq->scriptIn.str_, strlen(preq->scriptIn.str_),
                           "script", lnum, &rval);
  }
  if(!rv) {
    preq->scriptStatus = sc_status_success; 
    printf("JS_EvaluateScript failed:%s\n", preq->scriptIn.str_);
  } else {
    /* write result back to preq->scriptOut */
    preq->scriptStatus = sc_status_success;
    if(!JSVAL_IS_VOID(rval)) {
      /* check the se_scriptRequest_ReturnJSON flag and convert as needed */
      JSString* jres = JS_ValueToString(pl->jscx, rval);
      status = sc_string_assign(&preq->scriptOut, JS_GetStringBytes(jres));
      sm_status_returnIfFailed(status);
    }
  }

  /* signal client this is ready now */
  status = sc_sem_post(&preq->complete);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}


sc_status sc_lase_executeRequest(sc_lase* pl, sc_scriptRequestPtrh* ptrh)
{
  sc_status status = sc_status_success;
  sc_scriptRequest* preq = NULL;

  if(ptrh == NULL)
    sm_status_returnError(status, sc_err_nullArgument);

  preq = sc_scriptRequestPtrh_get(ptrh);
  if(preq == NULL) 
    sm_status_returnError(status, sc_err_nullArgument);

  /* check if this was called from the lase svc thread.  If so, just
   * execute it, otherwise submit. */
  if(pthread_equal(pthread_self(), pl->svcThread)) {
    status = sc_lase_executeHere(pl, ptrh);
    sm_status_returnIfFailed(status);
  } else {  
    status = sc_lase_submitRequest(pl, ptrh);
    sm_status_returnIfFailed(status);
  }
  /* wait for submitted request to complete */
  status = sc_sem_wait(&preq->complete);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}


sc_status sc_lase_serviceThread(sc_lase* pl)
{ /* serviceThread: all JavaScript is executed in this thread */
  sc_status status = sc_status_success;
  sc_scriptRequestPtrh rptrh;
  int done = 0, rv=0;
  struct pollfd fda[1];
  char ch;

  status = sc_scriptRequestPtrh_construct(&rptrh, NULL);
  sm_status_returnIfFailed(status);

  while(!done)
  {
    fda[0].fd = pl->ctlFdS;
    fda[0].events = POLLIN | POLLHUP;
    fda[0].revents = 0;    
  
    rv = poll(fda, 1, -1);
    if(rv == 1) {
      /* control pipe is readable */
      read(pl->ctlFdS, &ch, 1);
      if(ch == 4) {
        done = 1;
        break;
      } else if(ch == 0) {
        /* Grab a request out of the script queue and service it. */
        pthread_mutex_lock(&pl->mux);
        status = sc_scriptRequestPtrhQueue_popFront(&pl->reqs, &rptrh);
        pthread_mutex_unlock(&pl->mux);
        sm_status_returnIfFailed(status);

        /* execute script request */
        status = sc_lase_executeHere(pl, &rptrh);
        sm_status_returnIfFailed(status);

        /* release request rptrh */
        status = sc_scriptRequestPtrh_assignNewRep(&rptrh, NULL);
        sm_status_returnIfFailed(status);

        /* GC if memory is low */
        JS_MaybeGC(pl->jscx);
      }
    } else done = 1;

    if(pl->state == se_laseStateStopping)
      done = 1;
  }
  
  status = sc_scriptRequestPtrh_destruct(&rptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sm_queueImpl(sc_scriptRequestPtrh);
