#include <stdlib.h>
#include <string.h>
#include <si_sys/defs.h>

#include <jsapi.h>

static JSClass globclass = {
  "global", JSCLASS_GLOBAL_FLAGS,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

static JSClass autoclass = {
  "AutomationClass", JSCLASS_HAS_PRIVATE,
  JS_PropertyStub, /* property add */
  JS_PropertyStub, /* property del */
  JS_PropertyStub, /* property get */
  JS_PropertyStub, /* property set */
  JS_EnumerateStub,
  JS_ResolveStub,
  JS_ConvertStub,
  JS_FinalizeStub,
  JSCLASS_NO_OPTIONAL_MEMBERS
};

JSBool js_simpleSwitch_getName(JSContext *cx, JSObject *obj, jsval idval, jsval *vp);
JSBool js_simpleSwitch_setName(JSContext *cx, JSObject *obj, jsval idval, jsval *vp);
JSBool js_simpleSwitch_getState(JSContext *cx, JSObject *obj, jsval idval, jsval *vp);
JSBool js_simpleSwitch_setState(JSContext *cx, JSObject *obj, jsval idval, jsval *vp);


static JSPropertySpec wsps[] = {
  {"state", 0, JSPROP_ENUMERATE, js_simpleSwitch_getState, js_simpleSwitch_setState},
  {"name",  0, JSPROP_ENUMERATE, js_simpleSwitch_getName,  js_simpleSwitch_setName},
  {0}
};

typedef struct sc_simpleSwitch {
  const TCHAR* name;
  s_int32 state;
  void* jso;
} sc_simpleSwitch;

sc_status sc_simpleSwitch_construct(sc_simpleSwitch* psw, const TCHAR* name)
{
  sc_status status = sc_status_success;

  psw->jso   = NULL;
  psw->name  = NULL;
  psw->state = 0;

  status = sc_string_allocCopy((TCHAR**)&psw->name, name);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}
sc_status sc_simpleSwitch_destruct(sc_simpleSwitch* psw)
{
  sc_status status = sc_status_success;
  if(psw->name)
    free((void*)psw->name);  
  sm_status_return(status);
}
sc_status sc_simpleSwitch_getName(sc_simpleSwitch* psw, const TCHAR** name)
{
  sc_status status = sc_status_success;
  status = sc_string_allocCopy((TCHAR**)name, psw->name);
  sm_status_returnIfFailed(status);  
  sm_status_return(status);
}
sc_status sc_simpleSwitch_setName(sc_simpleSwitch* psw, const TCHAR* name)
{
  sc_status status = sc_status_success;
  status = sc_string_allocCopy((TCHAR**)&psw->name, name);
  sm_status_returnIfFailed(status);
  sm_status_return(status);
}
sc_status sc_simpleSwitch_getState(sc_simpleSwitch* psw, int* ps)
{
  *ps = psw->state;
  return sc_status_success;
}
sc_status sc_simpleSwitch_setState(sc_simpleSwitch* psw, int state)
{
  psw->state = state;
  return sc_status_success;
}


JSBool js_simpleSwitch_getName(JSContext *cx, JSObject *obj, jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_simpleSwitch* psw = NULL;  
  psw = (sc_simpleSwitch*) JS_GetInstancePrivate(cx, obj, &autoclass, NULL);
  
  if(psw) {
    char* str = NULL;
    status = sc_simpleSwitch_getName(psw, (const char**)&str);    
    *vp = STRING_TO_JSVAL(str);
    printf("js_simpleSwitch_getName(obj) == %s\n", str);
  }
  
  return JS_TRUE;
}
JSBool js_simpleSwitch_setName(JSContext *cx, JSObject *obj, jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_simpleSwitch* psw = NULL;  
  psw = (sc_simpleSwitch*) JS_GetInstancePrivate(cx, obj, &autoclass, NULL);
  
  if(psw) {
    status = sc_simpleSwitch_setName(psw, (char*)JSVAL_TO_STRING(*vp));
    printf("js_simpleSwitch_setName(obj, %s)\n", psw->name);
  }

  return JS_TRUE;
}
JSBool js_simpleSwitch_getState(JSContext *cx, JSObject *obj, jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_simpleSwitch* psw = NULL;  
  psw = (sc_simpleSwitch*) JS_GetInstancePrivate(cx, obj, &autoclass, NULL);

  if(psw) {
    int state;
    status = sc_simpleSwitch_getState(psw, &state);
    *vp = INT_TO_JSVAL(state);
    printf("js_simpleSwitch_getState(%s) == %d\n", psw->name, state);
  }

  return JS_TRUE;
}
JSBool js_simpleSwitch_setState(JSContext *cx, JSObject *obj, jsval idval, jsval *vp)
{
  sc_status status = sc_status_success;
  sc_simpleSwitch* psw = NULL;  
  psw = (sc_simpleSwitch*) JS_GetInstancePrivate(cx, obj, &autoclass, NULL);

  if(psw) {
    int state = JSVAL_TO_INT(*vp);
    status = sc_simpleSwitch_setState(psw, state);
    printf("js_simpleSwitch_setState(%s, %d);\n", psw->name, state);
  }

  return JS_TRUE;
}




/* The error reporter callback. Err, until I know what to do here */
void reportError(JSContext *cx, const char *message, JSErrorReport *report)
{
  fprintf(stderr, "%s:%u:%s\n",
          report->filename ? report->filename : "<no filename>",
          (unsigned int) report->lineno,
          message);
}

int main(int argc, char* argv[])
{
  sc_status status = sc_status_success;  
  /* script engine variables */
  JSRuntime *jsrt = NULL;
  JSContext *jscx = NULL;
  JSObject  *glob = NULL;
  JSObject  *wswo = NULL;
  sc_simpleSwitch* psw = NULL;
  JSBool    rv;
  jsval     rval;
  
  const char* fn = "script";
  int ln = 0;
  const char* script = "counter = 0;"
                       "while(counter < 10) { "
                       "  counter++; "
                       "  wallSwitch.state = (wallSwitch.state == 1) ? 0 : 1;"
                       "}";

  /* new runtime with an 8MB memory area */
  jsrt = JS_NewRuntime(8 * 1024 * 1024);
  if(jsrt == NULL) {
    perror("Unable to create JavaScript runtime: ");
    return 1;
  }
  jscx = JS_NewContext(jsrt, 8192);
  if(jscx == NULL) {
    perror("Unable to create JavaScript context: ");
    return 1;
  }

  /* set options */
  JS_SetOptions(jscx, JSOPTION_VAROBJFIX);
  JS_SetVersion(jscx, JSVERSION_LATEST);
  JS_SetErrorReporter(jscx, reportError);

  glob = JS_NewObject(jscx, &globclass, NULL, NULL);
  if(glob == NULL) {
    perror("Unable to create JavaScript global: ");
    return 1;
  }

  if(!JS_InitStandardClasses(jscx, glob)) {
    perror("JS_InitStandardClasses failed:");
    return 1;
  }

  /* create C object */
  psw = (sc_simpleSwitch*) malloc(sizeof(*psw));
  if(psw == NULL) {
    perror("Unable to create wall switch object");
    return 1;
  }  
  status = sc_simpleSwitch_construct(psw, "wallSwitch");
  sm_status_returnIfFailed(status);

  /* create wrapper */
  wswo = JS_DefineObject(jscx, glob, psw->name, &autoclass, NULL,
                         JSPROP_READONLY | JSPROP_ENUMERATE);
  if(wswo == NULL) {
    perror("Unable to create JavaScript wall switch object: ");
    return 1;
  }
  /* attach it to C object */
  rv = JS_SetPrivate(jscx, wswo, psw);
  if(!rv) {
    perror("JS_SetPrivateData failed.");
    return 1;
  }
  /* add the defined set of properties */
  rv = JS_DefineProperties(jscx, wswo, wsps); 
  if(!rv) {
    perror("JS_DefineProperties failed.");
    return 1;
  }
      

  /*
   * do some crazy stuff, like JS_EvaluateScript()
   */
  rv = JS_EvaluateScript(jscx, glob, script, strlen(script), fn, ln, &rval);
  if(!rv) {
    perror("JS_EvaluateScript failed.");
  }
  
  
  JS_MaybeGC(jscx);
  
  JS_DestroyContext(jscx);
  JS_DestroyRuntime(jsrt);
  JS_ShutDown();
  
  free(psw);
  return 0;
}
