#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <assert.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>

#include <si_sha/abrgdev.h>

#define scmax(a,b) (((a)>(b)) ? (a) : (b))


static void* serviceThreadEntry(void* arg);

sc_status sc_automationBridgeDevice_construct(
                            sc_automationBridgeDevice* pb,
                            sc_automationDeviceMgr*    pdm,
                            const TCHAR* name,
                            const TCHAR* type,
                            const TCHAR* addr,
                            const TCHAR* ntyp,
                            const sc_automationBridgeDevice_vtable* vtbl,
                            void* instance)
{
  sc_status status = sc_status_success;
  int ctl[2] = {-1, -1};

  pb->instance_ = instance;
  pb->vtbl_ = vtbl;
  pb->pdm_ = pdm;
  pb->state_ = se_bridgeStateStopped;

  pthread_mutex_init(&pb->mux_, NULL);
  sc_automationPacketPtrh_construct(&pb->xchg_, NULL);
  
  status = sc_automationDevice_construct(&pb->base_,
                                         name, type, addr, ntyp);
  sm_status_returnIfFailed(status);
  
  socketpair(AF_UNIX, SOCK_STREAM, 0, ctl);
  pb->fdCtlC_ = ctl[0];
  pb->fdCtlS_ = ctl[1];

  pb->state_ = se_bridgeStateStarting;
  /* auto start the device  */
  pthread_create(&pb->svcThread_, NULL, serviceThreadEntry, pb);

  
  sm_status_return(status);
}

sc_status sc_automationBridgeDevice_destruct(sc_automationBridgeDevice* pb)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 0x04; /* ^D */

  /* stop service thread */
  pb->state_ = se_bridgeStateStopping;
  write(pb->fdCtlC_, &ch, 1);
  pthread_join(pb->svcThread_, NULL);
  pb->state_ = se_bridgeStateStopped;
  
  /* bridge device destructor lines: */  
  status = sc_automationDevice_destruct(&pb->base_);
  sm_status_returnIfFailed(status);
    
  sm_status_return(status);
}


static
sc_status sc_automationBridgeDevice_serviceThread(sc_automationBridgeDevice* pb)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrh rqptrh, rsptrh;
  sc_automationPacket *preq, *pres;
  int rv, nfds;
  fd_set rfds;
  s_uint8 ch = 0x0e, in;
  
  /* listens to fd_ for async events, plus  */
  status = sc_automationPacketPtrh_construct(&rqptrh, NULL);
  sm_status_returnIfFailed(status);  
  status = sc_automationPacketPtrh_construct(&rsptrh, NULL);
  sm_status_returnIfFailed(status);

  nfds = scmax(pb->fd_, pb->fdCtlS_);
  nfds++;

  while(1)
  {
    FD_ZERO(&rfds);
    FD_SET(pb->fd_, &rfds);
    FD_SET(pb->fdCtlS_, &rfds);

    rv = select(nfds, &rfds, NULL, NULL, NULL);
    if(rv < 0) {
      status = sc_status_failure; 
      break;
    }
    if(pb->state_ == se_bridgeStateStopping)
      break;

    if(FD_ISSET(pb->fd_, &rfds)) {
      /* event from bridge */
      status = pb->vtbl_->newPacket(pb->instance_, &rsptrh);
      if(sm_status_failed(status))
        break;
      pres = sc_automationPacketPtrh_get(&rsptrh);
      
      status = pb->vtbl_->processEvent(pb->instance_, pres);
      if(sm_status_failed(status)) {
        /* ignore this for now */
      }
      /* queue event off to main thread */
    }
    if(FD_ISSET(pb->fdCtlS_, &rfds)) {
      /* new request from JavaScript */
      read(pb->fdCtlS_, &in, 1);

      if(in == 0x0f) {
        /* grab request */
        sc_automationPacketPtrh_assign(&rqptrh, &pb->xchg_);      
        /* allocate response */
        status = pb->vtbl_->newPacket(pb->instance_, &rsptrh);
        if(sm_status_failed(status))
          break;
        preq = sc_automationPacketPtrh_get(&rqptrh);
        pres = sc_automationPacketPtrh_get(&rsptrh);

        /* process by derived class allows identification of an
         * interleaved event before response */
        status = pb->vtbl_->processRequest(pb->instance_, preq, pres);
        if(sm_status_failed(status)) {
          /* ignore this for now */
        }

        sc_automationPacketPtrh_assign(&pb->xchg_, &rsptrh);
        write(pb->fdCtlS_, &ch, 1);
      } else if(in == 'd') {
        status = pb->vtbl_->discover(pb->instance_);
        if(sm_status_failed(status)) {
          /* ignore this for now */
        }
        in = 0; /* success */
        write(pb->fdCtlS_, &in, 1);
      }
    }
    
    /* listens to fd_ for async events, plus  */
  }

  status = sc_automationPacketPtrh_destruct(&rqptrh);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_destruct(&rsptrh);
  sm_status_returnIfFailed(status);  

  sm_status_return(status);
}

static void* serviceThreadEntry(void* arg)
{
  sc_automationBridgeDevice* pb = (sc_automationBridgeDevice*)arg;
  sc_status status = sc_status_success;
  status = sc_automationBridgeDevice_serviceThread(pb);
  assert(status == sc_status_success);
  return NULL;
}


sc_status sc_automationBridgeDevice_sendMessage(sc_automationBridgeDevice* brg,
                                                sc_automationPacket*       req,
                                                sc_automationPacketPtrh*   res)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 0x0f;
  sc_automationPacket* pres;
  
    
  pthread_mutex_lock(&brg->mux_);

  /* put the request at the exchange point */
  sc_automationPacketPtrh_assignNewRep(&brg->xchg_, &req->_rep);

  /* signal service thread that there is a request waiting */
  write(brg->fdCtlC_, &ch, 1);  

  /* wait for the response to come back */
  read(brg->fdCtlC_, &ch, 1);

  sc_automationPacketPtrh_assign(res, &brg->xchg_);
  
  pthread_mutex_unlock(&brg->mux_);
  
  sm_status_return(status);
}

sc_status sc_automationBridgeDevice_discover(sc_automationBridgeDevice* pbd)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 'd';

  pthread_mutex_lock(&pbd->mux_);

  /* signal service thread that a discovery is waiting */
  write(pbd->fdCtlC_, &ch, 1);  

  /* wait for the response to come back */
  read(pbd->fdCtlC_, &ch, 1);

  pthread_mutex_unlock(&pbd->mux_);
    
  sm_status_return(status);
}

sc_status sc_automationBridgeDevice_newPacket(sc_automationBridgeDevice* pbd,
                                              sc_automationPacketPtrh* pptrh)
{
  sc_status status = sc_status_success;
  
  status = pbd->vtbl_->newPacket(pbd->instance_, pptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}


