#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>


#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <si_std/string.h>
#include <si_sha/autobrg.h>

#include <assert.h>

#define INVALID_SOCKET (-1)
#define scmax(a,b) (((a)>(b)) ? (a) : (b))

sc_status sc_automationBridgeBase_serviceDevice(sc_automationBridgeBase* pb,
                                                sc_automationPacket* pkt);

/* unix socket names, install path, device metadata files, etc */

sc_status sc_automationBridgeBase_construct(
                                  sc_automationBridgeBase* pb,
                                  const TCHAR* name,
                                  const TCHAR* basePath,          
                                  s_uint32 pktSz, s_uint32 poolSz,
                                  const sc_automationBridge_vtable* vtbl,
                                  void* instance)
{
  sc_status status = sc_status_success;
  int ctl[2] = { -1, -1 };
  pb->instance_  = instance;
  pb->vtbl_      = vtbl;
  pb->svcThread_ = (pthread_t)-1;
  pb->state_     = se_bridgeStateStopped;

  status = sc_automationPacketPtrhList_construct(&pb->ooo_, 4, INT_MAX);
  sm_status_returnIfFailed(status);

  pthread_mutex_init(&pb->mux_, NULL);

  /* construct driver side pool (for events) as specified by derived
   * class */
  status = sc_automationPacketPool_construct(&pb->pool_,
                                             pktSz, poolSz, INT_MAX);
  sm_status_returnIfFailed(status);

  status = sc_automationBridgeConnectionPtrhList_construct(
                                             &pb->connections_,2,4);
  if(sm_status_failed(status)) {
    sc_automationPacketPool_destruct(&pb->pool_);
    sm_status_return(status);
  }

  pb->name_ = strdup(name);
  pb->basePath_ = strdup(basePath);
  
  sm_status_return(status);
}

static
sc_status sc_automationBridgeBase_serviceThread(sc_automationBridgeBase* pb)
{
  sc_status status = sc_status_success;
  fd_set rfds;
  struct timeval tv;
  int rv, nfds;
  sc_automationBridgeConnectionPtrh conn;
  sc_automationBridgeConnection* pnc = NULL;

  nfds = scmax(pb->fdCtlR_, pb->fdSvc_);
  nfds++;
  
  pb->state_ = se_bridgeStateStarted;

  while(1)
  {
    FD_ZERO(&rfds);
    FD_SET(pb->fdSvc_,  &rfds);
    FD_SET(pb->fdCtlR_, &rfds);
    
    tv.tv_sec = 5;
    tv.tv_usec = 0;

    rv = select(nfds, &rfds, NULL, NULL, &tv);
    if(rv > 0)
    {
      if(pb->state_ == se_bridgeStateStopping)
        break;
      if(FD_ISSET(pb->fdSvc_, &rfds)) {
        /* accept new client */
        int cs = accept(pb->fdSvc_, NULL, NULL);
        if(cs == INVALID_SOCKET)
          break;

        status = sc_automationBridgeConnectionPtrh_construct(&conn, NULL);
        if(sm_status_failed(status)) {
          close(cs);
          break;
        }
        /* create a new connection for accepted client */
        status = sc_automationBridgeConnection_new(pb, cs, &conn);
        if(sm_status_failed(status)) {
          close(cs);
          break;
        }

        /* add to connection list */
        pthread_mutex_lock(&pb->mux_);
        status = sc_automationBridgeConnectionPtrhList_pushBack(
                                             &pb->connections_, &conn);
        pthread_mutex_unlock(&pb->mux_);
        if(sm_status_failed(status)) {
          sc_automationBridgeConnectionPtrh_destruct(&conn);          
          break;
        }
        
        /* open the connection */
        pnc = sc_automationBridgeConnectionPtrh_get(&conn);        
        status = sc_automationBridgeConnection_open(pnc);
        if(sm_status_failed(status)) {
          sc_automationBridgeConnectionPtrh_destruct(&conn);          
          break;
        }        
        
        sc_automationBridgeConnectionPtrh_destruct(&conn);
      }
      if(FD_ISSET(pb->fdCtlR_, &rfds)) {
        /* could treat this as a way to take commands, not used */
      }
    }
    if(pb->state_ == se_bridgeStateStopping)
      break;    
  }

  sm_status_return(status);
}

static void* serviceThreadEntry(void* arg)
{ 
  sc_automationBridgeBase* pb = (sc_automationBridgeBase*)arg;
  sc_status status = sc_status_success;
  status = sc_automationBridgeBase_serviceThread(pb);
  assert(status == sc_status_success);
  return NULL;
}

sc_status sc_automationBridgeBase_start(sc_automationBridgeBase* pb)
{
  sc_status status = sc_status_success;
  int rv, servSock, ctl[2] = {-1, -1};
  struct sockaddr_un addr;
  sc_string namePath;
  mode_t mode;
  
  if(pb->state_ != se_bridgeStateStopped)
    sm_status_returnError(status, sc_status_failure);

  /* allocate the control pipe */
  rv = pipe(ctl);
  if(rv == -1)
    sm_status_returnError(status, sc_status_failure);

  pb->fdCtlR_ = ctl[0];
  pb->fdCtlW_ = ctl[1];

  /* create the service socket */
  servSock = socket(AF_UNIX, SOCK_STREAM, 0);
  if(servSock == INVALID_SOCKET) {
    close(ctl[0]); close(ctl[1]);
    sm_status_returnError(status, sc_status_failure);
  }

  memset(&addr, 0, sizeof(addr));
  addr.sun_family = AF_UNIX;

  /* construct service name */
  status = sc_string_constructFrom(&namePath, pb->basePath_);
  if(sm_status_failed(status)) {
    close(ctl[0]); close(ctl[1]); close(servSock);
    sm_status_return(status);
  }
  status = sc_string_append(&namePath, "/");
  if(sm_status_failed(status)) {
    close(ctl[0]); close(ctl[1]); close(servSock); sc_string_destruct(&namePath);
    sm_status_return(status);  
  }
  status = sc_string_append(&namePath, pb->name_);
  if(sm_status_failed(status)) {
    close(ctl[0]); close(ctl[1]); close(servSock); sc_string_destruct(&namePath);
    sm_status_return(status);
  }  

  /* unlink any previous entry with my name*/
  unlink(namePath.str_);  
  strncpy(addr.sun_path, namePath.str_, sizeof(addr.sun_path)-1);

  /* done with namePath temp */
  status = sc_string_destruct(&namePath);
  if(sm_status_failed(status)) {
    close(ctl[0]); close(ctl[1]); close(servSock); 
    sm_status_return(status);  
  }  

  /* name the service socket */  
  rv = bind(servSock, (struct sockaddr*)&addr, sizeof(addr));
  if(rv == -1) {
    close(ctl[0]); close(ctl[1]); close(servSock);
    sm_status_returnError(status, sc_status_failure);
  }

  /* put it in listen mode */
  rv = listen(servSock, 5);
  if(rv == -1) {
    close(ctl[0]); close(ctl[1]); close(servSock);
    sm_status_returnError(status, sc_status_failure);
  }

  /* this isn't the ideal permission for the socket, but I want to make 
   * it easy for anyone to run, at least for now */
  mode = S_IRUSR | S_IWUSR | S_IXUSR |
         S_IRGRP | S_IWGRP | S_IXGRP |
         S_IROTH | S_IWOTH | S_IXOTH;
  chmod(addr.sun_path, mode);
  
  pb->fdSvc_ = servSock;
  pb->state_ = se_bridgeStateStarting;
  pthread_create(&pb->svcThread_, NULL, serviceThreadEntry, pb);
  
  sm_status_return(status);
}

sc_status sc_automationBridgeBase_stop(sc_automationBridgeBase* pb)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 0x04; /* ^D */

  if(pb->state_ != se_bridgeStateStarted)
    sm_status_return(status);

  pb->state_ = se_bridgeStateStopping;

  /* wake 'em up */
  write(pb->fdCtlW_, &ch, 1);
  pthread_join(pb->svcThread_, NULL);

  /* control service socket and control pipes */
  close(pb->fdSvc_);
  close(pb->fdCtlW_);
  close(pb->fdCtlR_);

  pb->state_ = se_bridgeStateStopped;

  sm_status_return(status);
}

sc_status sc_automationBridgeBase_destruct(sc_automationBridgeBase* pb)
{
  sc_status status = sc_status_success;

  if(pb->state_ == se_bridgeStateStarted) {
    status = sc_automationBridgeBase_stop(pb);
    sm_status_returnIfFailed(status);
  }

  free((void*)pb->name_);
  free((void*)pb->basePath_);
  /* fdDev_ is closed by derived class */

  status = sc_automationPacketPtrhList_destruct(&pb->ooo_);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}


sc_status sc_automationBridgeBase_sendAsyncList(sc_automationBridgeBase* pb,
                                              sc_automationBridgeConnection* pc)
{
  sc_status status = sc_status_success;
  sc_automationPacketPtrhListIterator it;
  sc_automationPacket* pkt;

  /* should be re-sent on all connections, but multiple connections not
   * fully implemented */
  
  sc_automationPacketPtrhList_getBegin(&pb->ooo_, &it);  
  for(; !sc_automationPacketPtrhListIterator_isEnd(&it);
         sc_automationPacketPtrhListIterator_next(&it))
  {
    sc_automationPacketPtrh* pp = sc_automationPacketPtrhListIterator_get(&it);
    pkt = sc_automationPacketPtrh_get(pp);

    status = sc_automationPacket_serializeToFd(pkt, pc->fd_);
    sm_status_returnIfFailed(status);
  }

  status = sc_automationPacketPtrhList_clear(&pb->ooo_);
  sm_status_returnIfFailed(status);

  sm_status_return(status);  
}


sc_status sc_automationBridgeBase_processRequest(sc_automationBridgeBase* pb,
                                             sc_automationBridgeConnection* pc,
                                             sc_automationRequest* preq,
                                             sc_automationResponse* pres)
{
  sc_status status = sc_status_success;
  /* syntacted sugar over derived class call via v-table */
  pthread_mutex_lock(&pb->mux_);
  status = pb->vtbl_->processRequest(pb->instance_, pc, preq, pres);
  pthread_mutex_unlock(&pb->mux_);
  
  sm_status_return(status);
}

sc_status sc_automationBridgeBase_serviceDevice(sc_automationBridgeBase* pb,
                                                sc_automationPacket* pkt)
{
  sc_status status = sc_status_success;
  pthread_mutex_lock(&pb->mux_);
  status = pb->vtbl_->serviceDevice(pb->instance_, pkt);
  pthread_mutex_unlock(&pb->mux_);
  sm_status_return(status);
}

static 
sc_status sc_automationBridgeBase_removeConnection(sc_automationBridgeBase* pb,
                                              sc_automationBridgeConnection* pc)
{ /* remove and erase connection from active list */
  sc_status status = sc_status_success;
  sc_automationBridgeConnectionPtrhListIterator it;
  sc_automationBridgeConnectionPtrh* ptrh = NULL;
  sc_automationBridgeConnection* pitc = NULL;

  sc_automationBridgeConnectionPtrhListIterator_init(&it);

  /* lock connection list */
  pthread_mutex_lock(&pb->mux_); 
  
  status = sc_automationBridgeConnectionPtrhList_getBegin(&pb->connections_,&it);
  if(sm_status_failed(status)) {
    pthread_mutex_unlock(&pb->mux_);
    sm_status_return(status);
  }
  
  for(; !sc_automationBridgeConnectionPtrhListIterator_isEnd(&it);
         sc_automationBridgeConnectionPtrhListIterator_next(&it))
  {    
    ptrh = sc_automationBridgeConnectionPtrhListIterator_get(&it);
    pitc = sc_automationBridgeConnectionPtrh_get(ptrh);

    if(pc->fd_ == pitc->fd_) { /* found it, erase it */
      status = sc_automationBridgeConnectionPtrhList_erase(&pb->connections_,&it);
      break;
    }
  }

  /* release lock */
  pthread_mutex_unlock(&pb->mux_);  
  
  sm_status_return(status);
}


sc_status sc_automationBridgeBase_newPacket(sc_automationBridgeBase* pb,
                                            sc_automationPacketPtrh* pptrh)
{
  return sc_automationPacketPool_allocate(&pb->pool_, pptrh);
}

sc_status sc_automationBridgeBase_queueAsync(sc_automationBridgeBase* pb,
                                             sc_automationPacketPtrh* pptrh)
{
  sc_status status = sc_status_success;

  /* push it onto out-of-order list */
  status = sc_automationPacketPtrhList_pushBack(&pb->ooo_, pptrh);
  sm_status_returnIfFailed(status);
  
  sm_status_return(status);
}

sm_ptrhImpl(sc_automationBridgeConnection);
sm_listImpl(sc_automationBridgeConnectionPtrh);


sc_status sc_automationBridgeConnection_delete(void* pco, void* ponr)
{
  sc_status status = sc_status_success;
  sc_automationBridgeConnection* pc = (sc_automationBridgeConnection*)pco;

  status = sc_automationBridgeConnection_destruct(pc);
  sm_status_returnIfFailed(status);

  free(pc);

  sm_status_return(status);
}

sc_status sc_automationBridgeConnection_construct(
                                    sc_automationBridgeConnection* pc,
                                    sc_automationBridgeBase* pb, int fd)
{
  /* declare default Rep v-table */
  sm_repDefaultVtableDecl(vtbl,sc_automationBridgeConnection);
  sc_status status = sc_status_success;
  int rv, die[2] = {};
  
  pc->brg_  = pb;
  pc->open_ = seFalse;
  pc->fd_   = fd;

  status = sc_rep_construct(&pc->_rep, pc, NULL, &vtbl);
  sm_status_returnIfFailed(status);

  /* create die signal ... is there a better way to do this? */
  rv = pipe(die);
  if(rv == -1)
    sm_status_returnError(status, sc_status_failure);
  
  pc->dieR_ = die[0];
  pc->dieW_ = die[1];
  
  sm_status_return(status);
}


sc_status sc_automationBridgeConnection_new(sc_automationBridgeBase* pb, int fd,
                                    sc_automationBridgeConnectionPtrh* ptrh)
{
  sc_status status = sc_status_success;
  sc_automationBridgeConnection* pc;

  pc = (sc_automationBridgeConnection*) malloc(sizeof(sc_automationBridgeConnection));

  status = sc_automationBridgeConnection_construct(pc, pb, fd);
  if(sm_status_failed(status)) {
    free(pc);
    sm_status_return(status);
  }

  /* assign to out param */
  status = sc_automationBridgeConnectionPtrh_assignNewRep(ptrh, &pc->_rep);
  if(sm_status_failed(status)) {
    sc_automationBridgeConnection_delete(pc, NULL);
    sm_status_return(status);
  }
  
  sm_status_return(status);
}

static
sc_status sc_automationBridgeConnection_handle(
                                   sc_automationBridgeConnection* pc)
{ /* connection thread body */
  sc_status status = sc_status_success;
  sc_automationPacketPtrh pp0, pp1;
  sc_automationPacket* preq = NULL, *pres = NULL;  
  fd_set rfds;
  int rv, nfds;

  nfds = scmax(pc->fd_, pc->dieR_);
  nfds = scmax(pc->brg_->fdDev_, pc->dieR_);
  nfds++;

  /* allocate request & response packets */
  status = sc_automationPacketPtrh_construct(&pp0, NULL);
  sm_status_returnIfFailed(status);
  status = sc_automationPacketPtrh_construct(&pp1, NULL);
  sm_status_returnIfFailed(status);  
  status = sc_automationBridgeBase_newPacket(pc->brg_, &pp0);
  sm_status_returnIfFailed(status);
  status = sc_automationBridgeBase_newPacket(pc->brg_, &pp1);
  sm_status_returnIfFailed(status);
  preq = sc_automationPacketPtrh_get(&pp0);
  pres = sc_automationPacketPtrh_get(&pp1);
  
  while(1)
  {
    FD_ZERO(&rfds);
    FD_SET(pc->fd_, &rfds);
    FD_SET(pc->dieR_, &rfds);
    FD_SET(pc->brg_->fdDev_, &rfds);

    rv = select(nfds, &rfds, NULL, NULL, NULL);
    if(rv > 0) {
      if(FD_ISSET(pc->dieR_, &rfds)) 
        break; /* got a die signal */
      if(FD_ISSET(pc->brg_->fdDev_, &rfds)) {
        /* async event */
        status = sc_automationBridgeBase_serviceDevice(pc->brg_, pres);
        if(sm_status_failed(status))
          break;
        if(pres->priv_.size_ > 0) {
          status = sc_automationPacket_serializeToFd(pres, pc->fd_);
          if(sm_status_failed(status))
            break; /* other side closed? */
        }
      }
      if(FD_ISSET(pc->fd_, &rfds)) {
        /* command is readable, process it */
        status = sc_automationPacket_deserializeFromFd(preq, pc->fd_);
        if(sm_status_failed(status))
          break; /* other side closed? */

        /* calls down to derived class */
        status = sc_automationBridgeBase_processRequest(pc->brg_, pc, preq, pres);
        if(sm_status_failed(status))
          break; 

        if(pres->priv_.size_ > 0) {
          status = sc_automationPacket_serializeToFd(pres, pc->fd_);
          if(sm_status_failed(status))
            break; /* other side closed? */
        }
      }
      
      /* send off any OOO packets accumulated during this transaction. */
      sc_automationBridgeBase_sendAsyncList(pc->brg_, pc);
      /* clear packets for re-use  */
      sc_automationPacket_clear(preq);
      sc_automationPacket_clear(pres);
    } else break;
  }

  printf("bridge exiting %x\n", status);
  /* release local packets back to pool */
  sc_automationPacketPtrh_destruct(&pp0);
  sc_automationPacketPtrh_destruct(&pp1);  
  
  /* remove it from the bridge's connection list */
  sc_automationBridgeBase_removeConnection(pc->brg_, pc);

  sm_status_return(status);
}

static
void*  connectionThreadEntry(void* parg)
{
  sc_status status = sc_status_success;  
  sc_automationBridgeConnection* pc = ( sc_automationBridgeConnection*)parg;

  /* handle the connection */
  status =  sc_automationBridgeConnection_handle(pc);
  /* assert status succeeded */  
  
  return NULL;
}

sc_status sc_automationBridgeConnection_open(
                                          sc_automationBridgeConnection* pc)
{
  sc_status status = sc_status_success;

  if(!pc->open_) { /* create thread to handle connection */
    pc->open_ = seTrue;
    pthread_create(&pc->thread_, NULL, connectionThreadEntry, pc);
  }
  
  sm_status_return(status);
}

sc_status sc_automationBridgeConnection_close(
                                          sc_automationBridgeConnection* pc)
{
  sc_status status = sc_status_success;
  s_uint8 ch = 0x04;
  
  if(pc->open_) { /* signal connection to die and wait it for exit */
    write(pc->dieW_, &ch, 1);
    pthread_join(pc->thread_, NULL);
    pc->open_ = seFalse;
  }
  
  sm_status_return(status);
}

sc_status sc_automationBridgeConnection_destruct(
                                          sc_automationBridgeConnection* pc)
{
  sc_status status = sc_status_success;
  int selfDestruct = 0;

  /* If we are self-destructing, the connection is already closed. */
  selfDestruct = pc->open_ && pthread_equal(pthread_self(), pc->thread_);
  if(!selfDestruct) {
    status = sc_automationBridgeConnection_close(pc);
    sm_status_returnIfFailed(status);
  }
  
  close(pc->dieW_);
  close(pc->dieR_);
  close(pc->fd_);
  
  sm_status_return(status);
}
