#include <stdlib.h>
#include <string.h>
         
#include <si_sha/autodev.h>

sc_status sc_automationDevice_delete(void* obj, void* ownr);

sc_status sc_automationDevice_construct(sc_automationDevice* pd,
                                        const TCHAR* name,
                                        const TCHAR* type,
                                        const TCHAR* addr,
                                        const TCHAR* ntyp)
{
  sm_repDefaultVtableDecl(vtbl, sc_automationDevice);
  sc_status status = sc_status_success;

  pd->name_    = NULL;
  pd->type_    = NULL;
  pd->address_ = NULL;
  pd->netType_ = NULL;

  if(name)
    pd->name_    = strdup(name);
  if(type)
    pd->type_    = strdup(type);
  if(addr)
    pd->address_ = strdup(addr);
  if(ntyp)
    pd->netType_ = strdup(ntyp);

  status = sc_rep_construct(&pd->_rep, pd, NULL, &vtbl);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}

sc_status sc_automationDevice_destruct(sc_automationDevice* pd)
{
  sc_status status = sc_status_success;

  if(pd->name_)    free((void*)pd->name_);
  if(pd->type_)    free((void*)pd->type_);
  if(pd->address_) free((void*)pd->address_);
  if(pd->netType_) free((void*)pd->netType_);

  sm_status_return(status);
}

sc_status sc_automationDevice_constructCopy(sc_automationDevice* pd,
                                            sc_automationDevice* po)
{
  sc_status status = sc_status_success;

  pd->name_    = strdup(po->name_);
  pd->type_    = strdup(po->type_);
  pd->address_ = strdup(po->address_);
  pd->netType_ = strdup(po->netType_);

  sm_status_return(status);  
}
sc_status sc_automationDevice_assign(sc_automationDevice* pd,
                                     sc_automationDevice* po)
{
  sc_status status = sc_status_success;

  if(pd->name_)    free((void*)pd->name_);
  if(pd->type_)    free((void*)pd->type_);
  if(pd->address_) free((void*)pd->address_);
  if(pd->netType_) free((void*)pd->netType_);
  
  pd->name_    = strdup(po->name_);
  pd->type_    = strdup(po->type_);
  pd->address_ = strdup(po->address_);
  pd->netType_ = strdup(po->netType_);

  sm_status_return(status);  
}

/* instantiate device ptrh & ptrh list */
sm_ptrhImpl(sc_automationDevice);

sc_status sc_automationDevice_delete(void* obj, void* ownr)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd = (sc_automationDevice*)obj;
  
  status = sc_automationDevice_destruct(pd);
  sm_status_returnIfFailed(status);

  free(pd);
  
  sm_status_return(status);  
}

sc_status sc_automationDevice_new(const TCHAR* name,
                                  const TCHAR* type,
                                  const TCHAR* addr,
                                  const TCHAR* ntyp,
                                  sc_automationDevicePtrh* ptrh)
{
  sc_status status = sc_status_success;
  sc_automationDevice* pd;

  pd = (sc_automationDevice*) malloc(sizeof(pd[0]));

  status = sc_automationDevice_construct(pd, name, type, addr, ntyp);
  if(sm_status_failed(status)) {
    free(pd);
    sm_status_return(status);
  }

  sc_automationDevicePtrh_assignNewRep(ptrh, &pd->_rep);
  
  sm_status_return(status);
}


/*
 * do these once needed ...
sm_listImpl(sc_automationDevice);
sm_listImpl(sc_automationGroup);
*/
