#include <stdlib.h>

#include <si_std/ptrh.h>

sc_status sc_ptrh_construct(sc_ptrh* pptrh, sc_rep* prep)
{
  sc_status status = sc_status_success;

  if(prep) {
    status = prep->vtbl->addRef(prep);
    sm_status_returnIfFailed(status);
  }
  pptrh->priv_.rep_ = prep;
  return status;  
}
sc_status sc_ptrh_constructCopy(sc_ptrh* pptrh, sc_ptrh* pother)
{
  sc_status status = sc_status_success;
  if(pother && pother->priv_.rep_) {
    status = pother->priv_.rep_->vtbl->addRef(pother->priv_.rep_);
    sm_status_returnIfFailed(status);
    pptrh->priv_.rep_ = pother->priv_.rep_;
  } else
    pptrh->priv_.rep_ = NULL;
  return status;
}
sc_status sc_ptrh_assign(sc_ptrh* pptrh, sc_ptrh* prhs)
{
  sc_status status = sc_status_success;
  
  if(prhs && prhs->priv_.rep_) {
    status = prhs->priv_.rep_->vtbl->addRef(prhs->priv_.rep_);
    sm_status_returnIfFailed(status);
  }
  if(pptrh->priv_.rep_) {
    status = pptrh->priv_.rep_->vtbl->relRef(pptrh->priv_.rep_);
    sm_status_returnIfFailed(status);    
  }
  pptrh->priv_.rep_ = (prhs) ? prhs->priv_.rep_ : NULL;
  
  return status;
}
sc_status sc_ptrh_assignNewRep(sc_ptrh* pptrh, sc_rep* prep)
{
  sc_status status = sc_status_success;
  if(prep) {
    status = prep->vtbl->addRef(prep);
    sm_status_returnIfFailed(status);
  }
  if(pptrh->priv_.rep_) {
    status = pptrh->priv_.rep_->vtbl->relRef(pptrh->priv_.rep_);
    sm_status_returnIfFailed(status);    
  }
  pptrh->priv_.rep_ = prep;
  return status;
}
sc_status sc_ptrh_destruct(sc_ptrh* pptrh)
{
  sc_status status = sc_status_success;
  if(pptrh->priv_.rep_) {
    status = pptrh->priv_.rep_->vtbl->relRef(pptrh->priv_.rep_);
    sm_status_returnIfFailed(status);    
  }
  return status;
}


static __inline__ int atomic_inc(volatile int *v)
{
  int rv=1;
  asm volatile ("lock; xadd %0, %1"
                :"=r" (rv)
                :"m" (*v), "0" (rv));
  return rv+1;
}
static __inline__ int atomic_dec(volatile int *v)
{
  int rv=-1;
  asm volatile ("lock; xadd %0, %1"
                :"=r" (rv)
                :"m" (*v), "0" (rv));
  return rv-1;
}

sc_status sc_rep_addRef_default(sc_rep* prep)
{
  /* TODO: replace this with interlocked increment */
  atomic_inc(&prep->refc);
  return sc_status_success;
}
sc_status sc_rep_relRef_default(sc_rep* prep)
{  
  sc_status status = sc_status_success;
  /* TODO: replace this with interlocked decrement */
  if(atomic_dec(&prep->refc) == 0)
    return prep->vtbl->delete(prep->pobj, prep->pown);
  return status;
}

sc_status sc_rep_construct(sc_rep* rep, void* pobj, void* pown,
                           const sc_rep_vtable* vtbl)
{
  rep->pobj = pobj;
  rep->pown = pown;
  rep->vtbl = vtbl;
  rep->refc = 0;
  return sc_status_success;
}
