#include <stdlib.h>
#include <limits.h>

#include <si_std/queue.h>


/* macro for getting the value address in the queue node */
#define sm_valPtr(np) ((void*)(np+1))

sc_status sc_queue_construct(sc_queue* pq,
                             const sc_queueNodeDesc* pd,
                             s_int32 initialCapacity)
{
  sc_status status = sc_status_success;

  if(pq == NULL || pd == NULL)
    sm_status_returnError(status, sc_err_nullArgument);
  if(initialCapacity < 0)
    sm_status_returnError(status, sc_err_badArgument);
  
  pq->desc = pd;
  pq->head = NULL;
  pq->tail = NULL;
  
  status = sc_blockAllocator_construct(&pq->blka, pd->nodeSize + sizeof(sc_queueNode),
                                       initialCapacity, INT_MAX);
  sm_status_returnIfFailed(status);

  sm_status_return(status);
}


sc_status sc_queue_pushBack(sc_queue* pq, void* pnt)
{
  sc_status status = sc_status_success;
  sc_queueNode* pqn = NULL;

  status = sc_blockAllocator_allocate(&pq->blka, (void**)&pqn);
  sm_status_returnIfFailed(status);

  status = pq->desc->constructCopy(sm_valPtr(pqn), pnt);
  if(sm_status_failed(status)) {
    sc_blockAllocator_free(&pq->blka, pqn);    
    sm_status_return(status);
  }  
  
  pqn->next = NULL;
  if(pq->tail) 
    pq->tail->next = pqn;
  else
    pq->head = pqn;  
  pq->tail = pqn;
  
  sm_status_return(status);
}

sc_status sc_queue_popFront(sc_queue* pq, void* pnt)
{
  sc_status status = sc_status_success;
  sc_queueNode* pqn = NULL;

  if(pq == NULL)
    sm_status_returnError(status, sc_err_nullArgument);
  if(pq->head == NULL)
    sm_status_returnError(status, sc_err_badArgument); /* TODO: better error */

  pqn = pq->head;

  if(pnt != NULL)  {
    status = pq->desc->assign(pnt, sm_valPtr(pqn));
    sm_status_returnIfFailed(status);
  }
  
  pq->head = pqn->next;
  if(pq->head == NULL)
    pq->tail = NULL;

  pq->desc->destruct(sm_valPtr(pqn));
  sc_blockAllocator_free(&pq->blka, pqn);
  
  sm_status_return(status);
}

sc_status sc_queue_destruct(sc_queue* pq)
{
  sc_status status = sc_status_success;
  sc_queueNode* pn  = NULL;

  if(pq == NULL)
    sm_status_returnError(status, sc_err_nullArgument);
  while(pq->head != NULL) 
    sc_queue_popFront(pq, NULL);
  
  sm_status_return(status);
}
