#include <stdlib.h>
#include <string.h>
#include <si_std/string.h>


#define sm_strlen(s) strlen(s)
#define sm_strcpy(a,b) strcpy(a,b)


sc_status sc_string_construct(sc_string* pstr)
{
  pstr->str_ = NULL;
  pstr->capacity_ = 0;
  pstr->length_ = 0;
  return sc_status_success;
}

sc_status sc_string_constructFrom(sc_string* pstr,
                                  const TCHAR* src)
{
  pstr->str_ = NULL;
  pstr->capacity_ = 0;
  pstr->length_ = 0;  
  return sc_string_assign(pstr, src);
}

sc_status sc_string_destruct(sc_string* pstr)
{
  return sc_string_free(pstr);
}

sc_status sc_string_grow(sc_string* pstr, s_uint32 newcap) 
{ /* sets string capacity to newCap, reallocating and copying if needed */
  sc_status status = sc_status_success;
  TCHAR* pnstr=0;
  if(newcap > pstr->capacity_) {
    s_uint32 ncap = newcap * 3;
    ncap >>= 1;
    pnstr = (TCHAR*)malloc( (ncap+1) * sizeof(TCHAR));
    if(pnstr) {
      if(pstr->str_) {
        memcpy(pnstr, pstr->str_, pstr->length_ * sizeof(TCHAR));
        free(pstr->str_);
      }
      pstr->str_ = pnstr;
      pstr->capacity_ = ncap;
      *(pstr->str_+pstr->length_) = _T('\0');
    } else
      sm_status_returnError(status, sc_err_outOfMemory);
  }
  return sc_status_success;
}


sc_status sc_string_assign(sc_string* pstr, const TCHAR* src)
{	
  s_uint32 len;
  if(src == NULL) return sc_string_free(pstr);
  len = (s_uint32)sm_strlen(src);
  if(len == 0 && pstr->str_ == NULL) 
    return sc_string_grow(pstr, 1);	
  return sc_string_assignChars(pstr, src, len);
}

sc_status sc_string_assignChars(sc_string* pstr, const TCHAR* src, s_uint32 nchars)
{
  sc_status status = sc_status_success;
  if(src==0 && nchars==0) return sc_string_free(pstr);
  if(nchars > pstr->capacity_) {
    status = sc_string_grow(pstr,nchars);
    sm_status_returnIfFailed(status);
  }
  if(pstr->str_) {
    memcpy(pstr->str_, src, nchars * sizeof(TCHAR));
    pstr->length_ = nchars;
    *(pstr->str_+pstr->length_) = _T('\0');
  }
  sm_status_return(status);
}

sc_status sc_string_append(sc_string* pstr, const TCHAR* src)
{
  if(!src) return sc_status_success;
  return sc_string_appendChars(pstr, src, (s_uint32)sm_strlen(src));
}

sc_status sc_string_appendChars(sc_string* pstr, const TCHAR* src, s_uint32 nchars)
{
  sc_status status = sc_status_success;
  if( (!src) || (nchars==0) ) return 0;
  if(pstr->length_ + nchars > pstr->capacity_) {
    status = sc_string_grow(pstr, pstr->length_ + nchars );
    sm_status_returnIfFailed(status);
  }
  memcpy(pstr->str_+pstr->length_, src, nchars * sizeof(TCHAR));
  pstr->length_ += nchars;
  *(pstr->str_+pstr->length_) = _T('\0');

  sm_status_return(status);
}

sc_status sc_string_setMinCapacity(sc_string* pstr, s_uint32 numChars)
{
  return sc_string_grow(pstr, numChars);
}

sc_status sc_string_prepend(sc_string* pstr, const TCHAR* src)
{
  if(!src) return sc_status_success;
  return sc_string_prependChars(pstr, src, (s_uint32)strlen(src));  
}
sc_status sc_string_prependChars(sc_string* pstr, const TCHAR* src,
                                 s_uint32 nchars)
{
  sc_status status = sc_status_success;
  TCHAR *pch = NULL, *dst = NULL;
  int idx = 0, len;

  if( (!src) || (nchars == 0) )
    sm_status_return(status);

  if(pstr->length_ + nchars > pstr->capacity_) {
    status = sc_string_grow(pstr, pstr->length_ + nchars );
    sm_status_returnIfFailed(status);
  }

  /* copy old string backwards */
  len = pstr->length_;
  dst = pstr->str_ + len + nchars - 1;
  pch = pstr->str_ + len - 1;  
  while(idx < len) {
    *dst-- = *pch--;
    idx++;
  }  
  memcpy(pstr->str_, src, nchars * sizeof(TCHAR));

  pstr->length_ += nchars;
  *(pstr->str_+pstr->length_) = '\0';

  sm_status_return(status);  
}

sc_status sc_string_toLower(sc_string* pstr)
{
  sc_status status = sc_status_success;
  TCHAR* pch = NULL;
  int cidx = 0;

  pch = pstr->str_;
  for(cidx = 0; cidx < pstr->length_; ++cidx, ++pch)
  {
    TCHAR ch = *pch;
    if(ch >= _T('A') && ch <= _T('Z'))
      *pch = ch+32;
  }

  sm_status_return(status);
}

sc_status sc_string_free(sc_string* pstr)
{
  if(pstr->str_) 
    free(pstr->str_);
  pstr->str_ = 0;
  pstr->capacity_ = 0;
  pstr->length_ = 0;
  return 0;
}

sc_status sc_string_clear(sc_string* pstr)
{
  pstr->length_ = 0;
  if(pstr->str_)
    *(pstr->str_) = _T('\0');
  return sc_status_success;
}

sc_status sc_string_detach(sc_string* pstr, TCHAR** ppdst)
{
  sc_status status = sc_status_success;

  *ppdst = pstr->str_;
  pstr->str_ = NULL;
  pstr->length_ = 0;
  pstr->capacity_ = 0;

  sm_status_return(status);
}

sc_status sc_string_trimRight(sc_string* pstr)
{
  sc_status status = sc_status_success;

  if( (pstr == NULL) || (pstr->str_ == NULL)) {
    sm_status_return(status);
  } else {
    TCHAR* pch = pstr->str_ + pstr->length_ - 1;

    while (pstr->length_ && iswhitespace(*pch) ) {
      pstr->length_--;
      pch--;
    }  
    *(++pch) = _T('\0');
  }
  sm_status_return(status);  
}

sc_status sc_string_allocCopy(TCHAR** dst, const TCHAR* src)
{
  sc_status status = sc_status_success;
  TCHAR *newdst = *dst;

  if(newdst) free(newdst);
  newdst = NULL;

  if(src) {
    newdst = malloc( (sm_strlen(src) + 1) * sizeof(TCHAR));
    if(newdst == NULL)
      sm_status_returnError(status, sc_err_outOfMemory);
    sm_strcpy(newdst, src);
  }

  *dst = newdst;
  sm_status_return(status);
}


const TCHAR * str_advw(const TCHAR* psrc)
{
  TCHAR ch = *psrc;
  while(ch != '\0' && iswhitespace(ch)) 
    ch = *++psrc;
  return psrc;
}
const TCHAR* str_advq(const TCHAR* psrc)
{
  TCHAR ch = *psrc;
  while(ch != '\0' && ch != '\"')
    ch = *++psrc;
  return ++psrc;  
}
const TCHAR* str_advl(const TCHAR* psrc)
{
  TCHAR ch = *psrc;
  while(ch != _T('\0') && ch != _T('\r') && ch != _T('\n')) 
    ch = *++psrc;
  if(ch == _T('\r') && psrc[1] == _T('\n'))
    return psrc + 2;
  if(ch == _T('\0'))
    return psrc;
  return psrc + 1;
}
int iswhitespace(TCHAR c) {
  switch(c) {
    case _T(' ') :
    case _T('\r'):
    case _T('\n'):
    case _T('\t'):
      return 1;
    default:
      return 0;
  }
}

int _stricmp(const char* s1, 
             const char* s2)
{
  int c1, res;
  for (;;)
  {
    c1  =      *s1++;
    res = c1 - *s2++;
    if (c1 == 0) break;
    if (res != 0) {
      if(res ==  32 && c1 >= 'a' && c1 <= 'z') 
        continue;
      if(res == -32 && c1 >= 'A' && c1 <= 'Z')
        continue;
      break;
    }
  }
  return res;
}
