#include <stdlib.h>
#include <stdio.h>
#include <si_std/uuid.h>


sc_status sc_uuid_init(sc_uuid* pg)
{
  sc_status status = sc_status_success;
  s_int32 idx = 0;
  for(idx = 0; idx < 4; ++idx)
    pg->data[idx] = 0;
  sm_status_return(status);
}

sc_status sc_uuid_initFromString(sc_uuid* pg, const TCHAR* uuidstr)
{
  sc_status status = sc_status_success;
  s_uint8* u = &pg->data[0];
  int rv;  
  static const char* uuidFormat = "%2hhx%2hhx%2hhx%2hhx-"
                                  "%2hhx%2hhx-%2hhx%2hhx-%2hhx%2hhx-"
                                  "%2hhx%2hhx%2hhx%2hhx%2hhx%2hhx";

  /* sometimes you see the '{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}' format */ 
  if(uuidstr[0] == _T('{'))
    ++uuidstr; /* make only a minimal attempt to handle it */
  
  rv = sscanf(uuidstr, uuidFormat,
              &u[15], &u[14], &u[13], &u[12],
              &u[11], &u[10], &u[ 9], &u[ 8],
              &u[ 7], &u[ 6], &u[ 5], &u[ 4],
              &u[ 3], &u[ 2], &u[ 1], &u[ 0]);

  if(rv != 16)
    sm_status_returnError(status, sc_err_badArgument);
  
  sm_status_return(status);  
}

sc_status sc_uuid_assign(sc_uuid* pg, const sc_uuid* pog)
{
  sc_status status = sc_status_success;
  s_int32 idx;
  
  for(idx = 0; idx < 16; ++idx)
    pg->data[idx] = pog->data[idx];
  
  sm_status_return(status);
}

sc_status sc_uuid_isEqual(const sc_uuid* pThis, const sc_uuid* pThat,
                          s_bool* pAreEqual)
{
  sc_status status = sc_status_success;
  s_int32 idx;
  
  for(idx = 0; idx < 16; ++idx)
    if(pThis->data[idx] != pThat->data[idx])
      break;
  
  *pAreEqual = (idx == 16) ? seTrue : seFalse;    
  sm_status_return(status);
}

