#include <stdlib.h>
#include <si_sys/pthrdsem.h>

/* this was done as a simple version of sem_init that is portable with
 * pthreads, and therefore works under windows and Mac OS X  */

sc_status sc_sem_construct(sc_sem* s, s_int32 initialCount, s_int32 maxCount)
{
  sc_status status = sc_status_success;

  pthread_mutex_init(&s->m, NULL);
  pthread_cond_init(&s->c, NULL);
  s->count = initialCount;
  s->max = maxCount;

  sm_status_return(status);
}

sc_status sc_sem_wait(sc_sem* s)
{
  sc_status status = sc_status_success;  

  pthread_mutex_lock(&s->m);
  while(s->count == 0) {
    pthread_cond_wait(&s->c, &s->m);
  }
  s->count--;
  pthread_mutex_unlock(&s->m);
  
  sm_status_return(status);
}

sc_status sc_sem_trywait(sc_sem* s, s_bool* gotit)
{
  sc_status status = sc_status_success;  
  s_bool acquired = seFalse;
  
  pthread_mutex_lock(&s->m);
  if(s->count > 0) {
    s->count--;
    acquired = seTrue;
  }
  pthread_mutex_unlock(&s->m);

  if(gotit)
    *gotit = acquired;
  
  sm_status_return(status);  
}

sc_status sc_sem_post(sc_sem* s)
{
  sc_status status = sc_status_success;  

  pthread_mutex_lock(&s->m);
  if(s->count == s->max) {
    pthread_mutex_unlock(&s->m);    
    sm_status_returnError(status, sc_status_failure);
  }    
  if(++s->count == 1)
    pthread_cond_signal(&s->c);  
  pthread_mutex_unlock(&s->m);

  sm_status_return(status);
}

sc_status sc_sem_destruct(sc_sem* s)
{
  pthread_mutex_destroy(&s->m);
  pthread_cond_destroy(&s->c);
  return sc_status_success;
}


