#include <si_sys/serial.h>
#include <fcntl.h>



speed_t convertBaudRate(se_baudRate br)
{
  switch(br) {
    case se_baudRate_9600:
      return B9600;
    case se_baudRate_19200:
      return B19200;
    case se_baudRate_38400:
      return B38400;
    case se_baudRate_57600:
      return B57600;
    case se_baudRate_115200:
      return B115200;    
    default:
      return B19200;
  }
}

sc_status sc_serialPort_construct(sc_serialPort* pp, const char* dev,
                                  se_baudRate baud)
{
  sc_status status = sc_status_success;
  struct termios new_tcs;
  int rv;
  speed_t tiobaud = convertBaudRate(baud);

  pp->fd_ = open(dev, O_RDWR | O_NOCTTY | O_NDELAY);
  if(pp->fd_ == -1)
    sm_status_returnError(status, sc_status_failure);

  /* store old port settings */
  rv = tcgetattr(pp->fd_, &pp->ops_);
  if(rv == -1) {
    close(pp->fd_);
    sm_status_returnError(status, sc_status_failure);
  }

  new_tcs = pp->ops_;
  new_tcs.c_cflag = tiobaud | CS8 | CREAD | CLOCAL;
  new_tcs.c_lflag = 0;      /* no processing, no echo ... binary mode! */
  new_tcs.c_iflag = IGNPAR; /* ignore parity errors */
  new_tcs.c_oflag = 0;      /* no processing ... binary mode */

  new_tcs.c_cc[VTIME] = 0;  /* inter-char timer not used */
  new_tcs.c_cc[VMIN]  = 1;  /* 1 char min to satisfy read */

  tcflush(pp->fd_, TCIFLUSH);
  rv = tcsetattr(pp->fd_, TCSANOW, &new_tcs);  
  if(rv == -1) {
    close(pp->fd_);
    sm_status_returnError(status, sc_status_failure);
  }

  /* go back to blocking mode of the fd */
  rv = fcntl(pp->fd_, F_GETFL);
  fcntl(pp->fd_, F_SETFL, rv &~ FNDELAY); 
  tcflush(pp->fd_, TCIFLUSH);
  
  sm_status_return(status);
}


sc_status sc_serialPort_read(sc_serialPort* pp, s_uint8* buff, s_int32 len,
                             s_int32* bytesRead)
{
  sc_status status = sc_status_success;
  int br;
  
  br = read(pp->fd_, buff, len);
  *bytesRead = br;
  
  sm_status_return(status);
}

sc_status sc_serialPort_readAll(sc_serialPort* pp, s_uint8* buff, s_int32 len)
{
  sc_status status = sc_status_success;
  int br, tbr=0;

  while(tbr < len) {
    br = read(pp->fd_, buff+tbr, len-tbr);
    if(br <= 0)
      sm_status_returnError(status, sc_status_failure);
    tbr += br;
  }
  
  sm_status_return(status);
}

sc_status sc_serialPort_write(sc_serialPort* pp, const s_uint8* buff, 
                              s_int32 len, s_int32* bytesWritten)
{
  sc_status status = sc_status_success;
  int bw;
  
  bw = write(pp->fd_, buff, len);
  *bytesWritten = bw;
  
  sm_status_return(status);
}

sc_status sc_serialPort_flush(sc_serialPort* pp)
{
  sc_status status = sc_status_success;
  tcflush(pp->fd_, TCIFLUSH);
  sm_status_return(status);
}

sc_status sc_serialPort_destruct(sc_serialPort* pp)
{
  sc_status status = sc_status_success;

  tcsetattr(pp->fd_, TCSANOW, &pp->ops_);
  close(pp->fd_);
  
  sm_status_return(status);
}

/* should add 'readTimeout', 'waitForRead' */
